import pytest
import serpapi
from arcade_tdk.errors import ToolExecutionError

from arcade_search.utils import call_serpapi, prepare_params


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "engine, kwargs, expected",
    [
        ("google", {}, {"engine": "google"}),
        (
            "google",
            {"q": "test", "window": 10, "time": "00:12:12"},
            {
                "engine": "google",
                "q": "test",
                "window": 10,
                "time": "00:12:12",
            },
        ),
    ],
)
async def test_prepare_params(engine, kwargs, expected):
    params = prepare_params(engine, **kwargs)
    assert params == expected


@pytest.mark.parametrize(
    "error_message, sanitized_message",
    [
        (
            "You hit your rate limit",
            "You hit your rate limit",
        ),
        (
            "Bad Request for url: https://serpapi.com/search?engine=google_hotels&api_key=ABC123456",
            "Bad Request for url: https://serpapi.com/search?engine=google_hotels&api_key=***",
        ),
        (
            "Bad Request for url: https://serpapi.com/search?engine=google_hotels&api_key=ABC123456 make sure the api key is correct",
            "Bad Request for url: https://serpapi.com/search?engine=google_hotels&api_key=*** make sure the api key is correct",
        ),
    ],
)
def test_call_serpapi_failure(monkeypatch, dummy_context, error_message, sanitized_message):
    def fake_serpapi_search(self, params: dict) -> dict:
        raise Exception(error_message)  # noqa: TRY002

    monkeypatch.setattr(serpapi.Client, "search", fake_serpapi_search)

    with pytest.raises(ToolExecutionError) as excinfo:
        call_serpapi(dummy_context, {})

    assert excinfo.value.developer_message == sanitized_message



import pytest

from arcade_notion_toolkit.block_to_markdown_converter import BlockToMarkdownConverter


@pytest.mark.asyncio
async def test_convert_paragraph():
    block = {
        "type": "paragraph",
        "paragraph": {
            "rich_text": [
                {
                    "plain_text": "Hello, world!",
                    "annotations": {
                        "bold": False,
                        "italic": False,
                        "code": False,
                        "strikethrough": False,
                        "underline": False,
                        "color": "default",
                    },
                    "text": {"content": "Hello, world!", "link": None},
                    "type": "text",
                }
            ],
            "color": "default",
        },
    }
    converter = BlockToMarkdownConverter(context=None)
    result = await converter.convert_block(block)
    assert result == "Hello, world!  \n"


@pytest.mark.asyncio
async def test_convert_heading_1():
    block = {
        "type": "heading_1",
        "heading_1": {
            "rich_text": [
                {
                    "plain_text": "Heading Test",
                    "annotations": {
                        "bold": True,
                        "italic": False,
                        "code": False,
                        "strikethrough": False,
                        "underline": False,
                        "color": "default",
                    },
                    "text": {"content": "Heading Test", "link": None},
                    "type": "text",
                }
            ],
            "color": "default",
        },
    }
    converter = BlockToMarkdownConverter(context=None)
    result = await converter.convert_block(block)
    expected = "# **Heading Test**  \n"
    assert result == expected


@pytest.mark.asyncio
async def test_convert_bulleted_list_item():
    block = {
        "type": "bulleted_list_item",
        "bulleted_list_item": {
            "rich_text": [
                {
                    "plain_text": "list item",
                    "annotations": {
                        "bold": False,
                        "italic": False,
                        "code": False,
                        "strikethrough": False,
                        "underline": False,
                        "color": "default",
                    },
                    "text": {"content": "list item", "link": None},
                    "type": "text",
                }
            ],
            "color": "default",
        },
    }
    converter = BlockToMarkdownConverter(context=None)
    result = await converter.convert_block(block)
    expected = "- list item  \n"
    assert result == expected


@pytest.mark.asyncio
async def test_convert_equation():
    block = {"type": "equation", "equation": {"expression": "x+1=2"}}
    converter = BlockToMarkdownConverter(context=None)
    result = await converter.convert_block(block)
    expected = "$$ x+1=2 $$  \n"
    assert result == expected


@pytest.mark.asyncio
async def test_convert_child_page(monkeypatch):
    block = {
        "type": "child_page",
        "id": "child123",
        "child_page": {
            "rich_text": [
                {
                    "plain_text": "Child Title",
                    "annotations": {
                        "bold": False,
                        "italic": False,
                        "code": False,
                        "strikethrough": False,
                        "underline": False,
                        "color": "default",
                    },
                    "text": {"content": "Child Title", "link": None},
                    "type": "text",
                }
            ],
            "title": "Child Title",
        },
    }

    async def fake_get_page_url(context, block_id):
        return f"http://example.com/{block_id}"

    monkeypatch.setattr(
        "arcade_notion_toolkit.block_to_markdown_converter.get_page_url", fake_get_page_url
    )
    converter = BlockToMarkdownConverter(context=None)
    result = await converter.convert_block(block)
    expected = "[Child Title](http://example.com/child123)  \n"
    assert result == expected


@pytest.mark.asyncio
async def test_fallback_plaintext():
    block = {
        "type": "unsupported-type",
        "unsupported-type": {
            "rich_text": [
                {
                    "plain_text": "Fallback text",
                    "annotations": {
                        "bold": False,
                        "italic": False,
                        "code": False,
                        "strikethrough": False,
                        "underline": False,
                        "color": "default",
                    },
                    "text": {"content": "Fallback text", "link": None},
                    "type": "text",
                }
            ]
        },
    }
    converter = BlockToMarkdownConverter(context=None)
    result = await converter.convert_block(block)
    expected = "Fallback text"
    assert result == expected



from copy import deepcopy
from unittest.mock import AsyncMock, MagicMock, call, patch

import httpx
import pytest

from arcade_hubspot.enums import HubspotObject
from arcade_hubspot.exceptions import HubspotToolExecutionError, NotFoundError
from arcade_hubspot.models import HubspotCrmClient


@pytest.mark.asyncio
async def test_get_success(mock_context, mock_httpx_client):
    client = HubspotCrmClient(mock_context.get_auth_token_or_empty())

    mock_response = MagicMock(spec=httpx.Response)
    mock_response.status_code = 200
    mock_response.json.return_value = {"data": []}
    mock_httpx_client.get.return_value = mock_response

    response = await client.get("objects/contacts", params={"id": "123"})

    assert response == {"data": []}
    mock_httpx_client.get.assert_called_once_with(
        url="https://api.hubapi.com/crm/v3/objects/contacts",
        params={"id": "123"},
        headers={"Authorization": f"Bearer {mock_context.get_auth_token_or_empty()}"},
    )


@pytest.mark.asyncio
async def test_get_not_found_error(mock_context, mock_httpx_client):
    client = HubspotCrmClient(mock_context.get_auth_token_or_empty())

    mock_response = MagicMock(spec=httpx.Response)
    mock_response.status_code = 404
    mock_response.json.return_value = {"message": "Not Found", "errors": [{"message": "Not Found"}]}
    mock_httpx_client.get.return_value = mock_response

    with pytest.raises(NotFoundError) as exc_info:
        await client.get("objects/contacts", params={"id": "123"})

    assert exc_info.value.message == "Not Found"
    assert exc_info.value.developer_message == '[{"message": "Not Found"}]'


@pytest.mark.asyncio
async def test_get_internal_server_error(mock_context, mock_httpx_client):
    client = HubspotCrmClient(mock_context.get_auth_token_or_empty())

    mock_response = MagicMock(spec=httpx.Response)
    mock_response.status_code = 500
    mock_response.json.return_value = {
        "message": "Internal Server Error",
        "errors": [{"message": "Internal Server Error"}],
    }
    mock_httpx_client.get.return_value = mock_response

    with pytest.raises(HubspotToolExecutionError) as exc_info:
        await client.get("objects/contacts", params={"id": "123"})

    assert exc_info.value.message == "Internal Server Error"
    assert exc_info.value.developer_message == '[{"message": "Internal Server Error"}]'


@pytest.mark.asyncio
async def test_get_with_invalid_json_error_response(mock_context, mock_httpx_client):
    client = HubspotCrmClient(mock_context.get_auth_token_or_empty())

    mock_response = MagicMock(spec=httpx.Response)
    mock_response.status_code = 500
    mock_response.text = "Text error message"
    mock_response.json.side_effect = ValueError("Invalid JSON")
    mock_httpx_client.get.return_value = mock_response

    with pytest.raises(HubspotToolExecutionError) as exc_info:
        await client.get("objects/contacts", params={"id": "123"})

    assert exc_info.value.message == "Text error message"
    assert exc_info.value.developer_message is None


@pytest.mark.asyncio
async def test_post_success(mock_context, mock_httpx_client):
    client = HubspotCrmClient(mock_context.get_auth_token_or_empty())

    mock_response = MagicMock(spec=httpx.Response)
    mock_response.status_code = 200
    mock_response.json.return_value = {"data": []}
    mock_httpx_client.post.return_value = mock_response

    response = await client.post("objects/contacts", json_data={"id": "123"})

    assert response == {"data": []}
    mock_httpx_client.post.assert_called_once_with(
        url="https://api.hubapi.com/crm/v3/objects/contacts",
        json={"id": "123"},
        headers={
            "Authorization": f"Bearer {mock_context.get_auth_token_or_empty()}",
            "Content-Type": "application/json",
        },
    )


@pytest.mark.asyncio
async def test_post_not_found_error(mock_context, mock_httpx_client):
    client = HubspotCrmClient(mock_context.get_auth_token_or_empty())

    mock_response = MagicMock(spec=httpx.Response)
    mock_response.status_code = 404
    mock_response.json.return_value = {"message": "Not Found", "errors": [{"message": "Not Found"}]}
    mock_httpx_client.post.return_value = mock_response

    with pytest.raises(NotFoundError) as exc_info:
        await client.post("objects/contacts", json_data={"id": "123"})

    assert exc_info.value.message == "Not Found"
    assert exc_info.value.developer_message == '[{"message": "Not Found"}]'


@pytest.mark.asyncio
async def test_post_internal_server_error(mock_context, mock_httpx_client):
    client = HubspotCrmClient(mock_context.get_auth_token_or_empty())

    mock_response = MagicMock(spec=httpx.Response)
    mock_response.status_code = 500
    mock_response.json.return_value = {
        "message": "Internal Server Error",
        "errors": [{"message": "Internal Server Error"}],
    }
    mock_httpx_client.post.return_value = mock_response

    with pytest.raises(HubspotToolExecutionError) as exc_info:
        await client.post("objects/contacts", json_data={"id": "123"})

    assert exc_info.value.message == "Internal Server Error"
    assert exc_info.value.developer_message == '[{"message": "Internal Server Error"}]'


@pytest.mark.asyncio
@patch("arcade_hubspot.models.clean_data")
async def test_batch_get_objects(mock_clean_data, mock_context, mock_httpx_client):
    mock_results = [MagicMock(spec=dict) for _ in range(3)]

    clean_data_response = [MagicMock(spec=dict) for _ in range(3)]
    mock_clean_data.side_effect = clean_data_response

    mock_response = MagicMock(spec=httpx.Response)
    mock_response.status_code = 200
    mock_response.json.return_value = {
        "results": mock_results,
    }
    mock_httpx_client.post.return_value = mock_response

    client = HubspotCrmClient(mock_context.get_auth_token_or_empty())
    response = await client.batch_get_objects(
        HubspotObject.CONTACT, ["123", "456"], ["email", "firstname"]
    )

    assert response == clean_data_response
    mock_clean_data.assert_has_calls([
        call(result, HubspotObject.CONTACT) for result in mock_results
    ])

    mock_httpx_client.post.assert_called_once_with(
        url="https://api.hubapi.com/crm/v3/objects/contacts/batch/read",
        json={
            "inputs": [{"id": "123"}, {"id": "456"}],
            "properties": ["email", "firstname"],
        },
        headers={
            "Authorization": f"Bearer {mock_context.get_auth_token_or_empty()}",
            "Content-Type": "application/json",
        },
    )


@pytest.mark.asyncio
@patch("arcade_hubspot.models.clean_data")
async def test_get_associated_objects(mock_clean_data, mock_context, mock_httpx_client):
    # Mock stuff from the batch_get_objects function
    mock_batch_results = [MagicMock(spec=dict) for _ in range(3)]
    clean_data_response = [MagicMock(spec=dict) for _ in range(3)]
    mock_clean_data.side_effect = clean_data_response
    mock_post_response = MagicMock(spec=httpx.Response)
    mock_post_response.status_code = 200
    mock_post_response.json.return_value = {
        "results": mock_batch_results,
    }
    mock_httpx_client.post.return_value = mock_post_response

    # Mock stuff from the get_associated_objects function
    mock_response = MagicMock(spec=httpx.Response)
    mock_response.status_code = 200
    mock_response.json.return_value = {
        "results": [{"toObjectId": "123"}, {"toObjectId": "456"}],
    }
    mock_httpx_client.get.return_value = mock_response

    client = HubspotCrmClient(mock_context.get_auth_token_or_empty())
    response = await client.get_associated_objects(
        parent_object=HubspotObject.COMPANY,
        parent_id="parent_id",
        associated_object=HubspotObject.CONTACT,
        limit=5,
        after=5,
        properties=["email", "firstname"],
    )

    assert response == clean_data_response
    mock_clean_data.assert_has_calls([
        call(result, HubspotObject.CONTACT) for result in mock_batch_results
    ])

    mock_httpx_client.get.assert_called_once_with(
        url="https://api.hubapi.com/crm/v4/objects/company/parent_id/associations/contact",
        params={
            "limit": 5,
            "after": 5,
        },
        headers={
            "Authorization": f"Bearer {mock_context.get_auth_token_or_empty()}",
        },
    )

    mock_httpx_client.post.assert_called_once_with(
        url="https://api.hubapi.com/crm/v3/objects/contacts/batch/read",
        json={
            "inputs": [{"id": "123"}, {"id": "456"}],
            "properties": ["email", "firstname"],
        },
        headers={
            "Authorization": f"Bearer {mock_context.get_auth_token_or_empty()}",
            "Content-Type": "application/json",
        },
    )


@pytest.mark.asyncio
@patch("arcade_hubspot.models.prepare_api_search_response")
@patch("arcade_hubspot.models.get_object_properties")
async def test_search_by_keywords(
    mock_get_object_properties,
    mock_prepare_api_search_response,
    mock_context,
    mock_httpx_client,
):
    mock_response = MagicMock(spec=httpx.Response)
    mock_response.status_code = 200
    mock_httpx_client.post.return_value = mock_response

    mock_prepared_api_response = {
        "contacts": [
            {
                "id": "123",
                "first_name": "Jason",
                "last_name": "Bourne",
                "email": "jason.bourne@acme.com",
            }
        ]
    }
    mock_prepare_api_search_response.return_value = mock_prepared_api_response
    mock_get_object_properties.return_value = ["email", "firstname"]

    mock_calls = [{"id": f"call_{i}"} for i in range(3)]
    mock_emails = [{"id": f"email_{i}"} for i in range(3)]
    mock_notes = [{"id": f"note_{i}"} for i in range(3)]

    mock_get_associated_objects = AsyncMock()
    mock_get_associated_objects.side_effect = [mock_calls, mock_emails, mock_notes]

    client = HubspotCrmClient(mock_context.get_auth_token_or_empty())
    client.get_associated_objects = mock_get_associated_objects

    response = await client.search_by_keywords(
        object_type=HubspotObject.CONTACT,
        keywords="test",
        limit=10,
        associations=[
            HubspotObject.CALL,
            HubspotObject.EMAIL,
            HubspotObject.NOTE,
        ],
    )

    expected_response = deepcopy(mock_prepared_api_response)
    expected_response["contacts"][0]["calls"] = mock_calls
    expected_response["contacts"][0]["emails"] = mock_emails
    expected_response["contacts"][0]["notes"] = mock_notes

    assert response == expected_response

    mock_httpx_client.post.assert_called_once_with(
        url="https://api.hubapi.com/crm/v3/objects/contacts/search",
        headers={
            "Authorization": f"Bearer {mock_context.get_auth_token_or_empty()}",
            "Content-Type": "application/json",
        },
        json={
            "query": "test",
            "limit": 10,
            "sorts": [{"propertyName": "hs_lastmodifieddate", "direction": "DESCENDING"}],
            "properties": ["email", "firstname"],
        },
    )

    mock_get_associated_objects.assert_has_calls([
        call(
            parent_object=HubspotObject.CONTACT,
            parent_id="123",
            associated_object=HubspotObject.CALL,
            limit=10,
        ),
        call(
            parent_object=HubspotObject.CONTACT,
            parent_id="123",
            associated_object=HubspotObject.EMAIL,
            limit=10,
        ),
        call(
            parent_object=HubspotObject.CONTACT,
            parent_id="123",
            associated_object=HubspotObject.NOTE,
            limit=10,
        ),
    ])



from arcade_salesforce.tools.crm.account import (
    get_account_data_by_id,
    get_account_data_by_keywords,
)
from arcade_salesforce.tools.crm.contact import create_contact

__all__ = [
    "create_contact",
    "get_account_data_by_id",
    "get_account_data_by_keywords",
]



from unittest.mock import AsyncMock, patch

import pytest
from arcade_tdk.errors import ToolExecutionError
from httpx import Response

from arcade_github.tools.activity import list_stargazers, set_starred


@pytest.fixture
def mock_context():
    context = AsyncMock()
    context.authorization.token = "mock_token"  # noqa: S105
    return context


@pytest.fixture
def mock_client():
    with patch("arcade_github.tools.activity.httpx.AsyncClient") as client:
        yield client.return_value.__aenter__.return_value


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "starred,expected_message",
    [
        (True, "Successfully starred the repository owner/repo"),
        (False, "Successfully unstarred the repository owner/repo"),
    ],
)
async def test_set_starred_success(mock_context, mock_client, starred, expected_message):
    mock_client.put.return_value = mock_client.delete.return_value = Response(204)

    result = await set_starred(mock_context, "owner", "repo", starred)
    assert result == expected_message


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "status_code,error_message,expected_error",
    [
        (403, "Forbidden", "Error accessing.*: Forbidden"),
        (404, "Not Found", "Error accessing.*: Resource not found"),
        (500, "Internal Server Error", "Error accessing.*: Failed to process request"),
    ],
)
async def test_set_starred_errors(
    mock_context, mock_client, status_code, error_message, expected_error
):
    mock_client.put.return_value = mock_client.delete.return_value = Response(
        status_code, json={"message": error_message}
    )

    with pytest.raises(ToolExecutionError, match=expected_error):
        await set_starred(mock_context, "owner", "repo", True)


@pytest.mark.asyncio
async def test_list_stargazers_success(mock_context, mock_client):
    mock_response_data = [
        {
            "login": "user1",
            "id": 1,
            "node_id": "MDQ6VXNlcjE=",
            "html_url": "https://github.com/user1",
        },
        {
            "login": "user2",
            "id": 2,
            "node_id": "MDQ6VXNlcjI=",
            "html_url": "https://github.com/user2",
        },
    ]
    mock_client.get.return_value = Response(200, json=mock_response_data)

    result = await list_stargazers(mock_context, "owner", "repo", limit=2)
    assert result == {"number_of_stargazers": 2, "stargazers": mock_response_data}


@pytest.mark.asyncio
async def test_list_stargazers_empty(mock_context, mock_client):
    mock_client.get.return_value = Response(200, json=[])

    result = await list_stargazers(mock_context, "owner", "repo")
    assert result == {"number_of_stargazers": 0, "stargazers": []}


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "status_code,error_message,expected_error",
    [
        (403, "Forbidden", "Error accessing.*: Forbidden"),
        (404, "Not Found", "Error accessing.*: Resource not found"),
        (500, "Internal Server Error", "Error accessing.*: Failed to process request"),
    ],
)
async def test_list_stargazers_errors(
    mock_context, mock_client, status_code, error_message, expected_error
):
    mock_client.get.return_value = Response(status_code, json={"message": error_message})

    with pytest.raises(ToolExecutionError, match=expected_error):
        await list_stargazers(mock_context, "owner", "repo")



import httpx
import pytest
from arcade_tdk.errors import ToolExecutionError

from arcade_reddit.client import RedditClient
from arcade_reddit.utils import (
    create_fullname_for_comment,
    create_fullname_for_multiple_posts,
    create_fullname_for_post,
    create_path_for_post,
    parse_get_content_of_multiple_posts_response,
    parse_get_content_of_post_response,
    parse_get_posts_in_subreddit_response,
    parse_get_top_level_comments_response,
    parse_subreddit_rules_response,
    parse_user_posts_response,
    resolve_subreddit_access,
)


class DummyRedditClientResponse:
    def __init__(self, status_code):
        self.status_code = status_code


@pytest.mark.parametrize(
    "identifier, expected",
    [
        ("abcdef", "t1_abcdef"),
        ("https://www.reddit.com/r/test/comments/1234567890/comment/1abcdef/", "t1_1abcdef"),
        ("t1_abcdef", "t1_abcdef"),
        ("/r/test/comments/1234567890/comment/1abcdef/", "t1_1abcdef"),
        ("not-an-id", pytest.raises(ToolExecutionError)),
        # Fullname: Invalid (not a Reddit comment id, but a Reddit post id)
        ("t2_abcdef", pytest.raises(ToolExecutionError)),
        # URL: Invalid (not a Reddit url to a comment, but to a post)
        ("https://www.reddit.com/r/test/comments/1234567890", pytest.raises(ToolExecutionError)),
        # Permalink: Invalid (not a Reddit permalink to a comment, but to a post)
        ("/r/test/comments/1234567890", pytest.raises(ToolExecutionError)),
    ],
)
def test_create_fullname_for_comment(identifier, expected) -> None:
    if isinstance(expected, str):
        assert create_fullname_for_comment(identifier) == expected
    else:
        with expected:
            create_fullname_for_comment(identifier)


@pytest.mark.parametrize(
    "identifier, expected",
    [
        ("https://www.reddit.com/r/test/comments/1234abc/", "/r/test/comments/1234abc/"),
        ("1234abc", "/comments/1234abc"),
        ("/r/test/comments/1234abc/", "/r/test/comments/1234abc/"),
        ("t3_1234abc", "/comments/1234abc"),
        # URL: invalid (non-reddit domain)
        ("https://www.example.com/r/test/comments/1234abc/", pytest.raises(ToolExecutionError)),
        # Post ID: invalid (non-alphanumeric character)
        ("12!abc", pytest.raises(ToolExecutionError)),
        # Permalink: invalid (missing the leading "/" so not recognized as a proper permalink)
        ("r/test/comments/1234abc/", pytest.raises(ToolExecutionError)),
        # Fullname: invalid (contains an illegal character)
        ("t3_1234*abc", pytest.raises(ToolExecutionError)),
    ],
)
def test_create_path_for_post(identifier, expected):
    if isinstance(expected, str):
        result = create_path_for_post(identifier)
        assert result == expected
    else:
        with expected:
            create_path_for_post(identifier)


@pytest.mark.parametrize(
    "identifier, expected",
    [
        ("https://www.reddit.com/r/test/comments/1234abc/", "t3_1234abc"),
        ("1234abc", "t3_1234abc"),
        ("/r/test/comments/1234abc/", "t3_1234abc"),
        ("t3_1234abc", "t3_1234abc"),
        # URL: invalid (missing "/comments/" segment)
        ("https://www.reddit.com/r/test/1234abc/", pytest.raises(ToolExecutionError)),
        # Post ID: invalid (non-alphanumeric)
        ("12!abc", pytest.raises(ToolExecutionError)),
        # Permalink: invalid (missing "/comments/" segment)
        ("/r/test/1234abc/", pytest.raises(ToolExecutionError)),
        # Fullname: invalid (type prefix is for a message, not a post
        ("t4_1234abc", pytest.raises(ToolExecutionError)),
    ],
)
def test_create_fullname_for_post(identifier, expected):
    if isinstance(expected, str):
        result = create_fullname_for_post(identifier)
        assert result == expected
    else:
        with expected:
            create_fullname_for_post(identifier)


@pytest.mark.parametrize(
    "identifiers, expected_fullnames, expected_warnings",
    [
        ([], [], []),
        (
            ["t3_1234abc", "https://www.reddit.com/r/test/comments/1234abc/"],
            ["t3_1234abc", "t3_1234abc"],
            [],
        ),
        (
            ["t3_1234abc", "not-a-valid-identifier"],
            ["t3_1234abc"],
            [
                {
                    "message": "'not-a-valid-identifier' is not a valid Reddit post identifier.",
                    "identifier": "not-a-valid-identifier",
                }
            ],
        ),
        (
            ["inv@lid", "not-a-valid-identifier"],
            [],
            [
                {
                    "message": "'inv@lid' is not a valid Reddit post identifier.",
                    "identifier": "inv@lid",
                },
                {
                    "message": "'not-a-valid-identifier' is not a valid Reddit post identifier.",
                    "identifier": "not-a-valid-identifier",
                },
            ],
        ),
    ],
)
def test_create_fullname_for_multiple_posts(identifiers, expected_fullnames, expected_warnings):
    actual_fullnames, actual_warnings = create_fullname_for_multiple_posts(identifiers)
    assert actual_fullnames == expected_fullnames
    assert actual_warnings == expected_warnings


def test_parse_get_posts_in_subreddit_response_empty():
    data = {}
    expected = {"cursor": None, "posts": []}
    result = parse_get_posts_in_subreddit_response(data)
    assert result == expected


def test_parse_get_content_of_post_response_empty_and_malformed():
    data = []
    result = parse_get_content_of_post_response(data)
    assert result == {}

    data = None
    result = parse_get_content_of_post_response(data)
    assert result == {}

    # missing expected keys
    data = [{}]
    expected = {
        "id": None,
        "name": None,
        "title": None,
        "author": None,
        "subreddit": None,
        "created_utc": None,
        "num_comments": None,
        "score": None,
        "upvote_ratio": None,
        "upvotes": None,
        "permalink": None,
        "url": None,
        "is_video": None,
        "body": None,
    }
    result = parse_get_content_of_post_response(data)
    assert result == expected


def test_parse_get_content_of_multiple_posts_response_empty_and_malformed():
    expected = []

    data = {}
    result = parse_get_content_of_multiple_posts_response(data)
    assert result == expected

    data = None
    result = parse_get_content_of_multiple_posts_response(data)
    assert result == expected

    data = [{}]
    result = parse_get_content_of_multiple_posts_response(data)
    assert result == expected


def test_parse_get_top_level_comments_response_missing_data():
    data = [{}]
    expected = {"comments": [], "num_comments": 0}
    result = parse_get_top_level_comments_response(data)
    assert result == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "status_code, expected_result, expected_exception",
    [
        (None, {"exists": True, "accessible": True}, None),
        (404, {"exists": False, "accessible": False}, None),
        (302, {"exists": False, "accessible": False}, None),
        (403, {"exists": True, "accessible": False}, None),
        (500, None, httpx.HTTPStatusError),
    ],
)
async def test_resolve_subreddit_access(status_code, expected_result, expected_exception):
    class DummyRedditClient(RedditClient):
        async def get(self, path: str, **kwargs):
            if status_code is not None:
                raise httpx.HTTPStatusError(
                    "Error", request=None, response=DummyRedditClientResponse(status_code)
                )
            return "dummy_success"

    client = DummyRedditClient(token="dummy")  # noqa: S106
    if expected_exception:
        with pytest.raises(expected_exception):
            await resolve_subreddit_access(client, "testsub")
    else:
        result = await resolve_subreddit_access(client, "testsub")
        assert result == expected_result


@pytest.mark.parametrize(
    "data, expected",
    [
        (
            {
                "rules": [
                    {"priority": 1, "short_name": "Rule1", "description": "Desc1"},
                    {"priority": 2, "short_name": "Rule2", "description": "Desc2"},
                ]
            },
            {
                "rules": [
                    {"priority": 1, "title": "Rule1", "body": "Desc1"},
                    {"priority": 2, "title": "Rule2", "body": "Desc2"},
                ]
            },
        ),
        ({}, {"rules": []}),
    ],
)
def test_parse_subreddit_rules_response(data, expected):
    result = parse_subreddit_rules_response(data)
    assert result == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "posts_data, expected_cursor, expected_posts",
    [
        (
            {
                "data": {
                    "after": "cursor123",
                    "children": [
                        {
                            "data": {
                                "id": "1",
                                "name": "t3_1",
                                "title": "Post 1",
                                "author": "user1",
                                "subreddit": "subreddit1",
                                "created_utc": 1712345678,
                                "num_comments": 10,
                                "score": 100,
                                "upvote_ratio": 0.5,
                                "ups": 50,
                                "permalink": "permalink1",
                                "url": "url1",
                                "is_video": False,
                                "selftext": "This is the body of post 1",
                                "all_awardings": [],
                                "allow_live_comments": False,
                                "approved": False,
                                "approved_at_utc": None,
                            }
                        },
                    ],
                }
            },
            "cursor123",
            [
                {
                    "id": "1",
                    "name": "t3_1",
                    "title": "Post 1",
                    "author": "user1",
                    "subreddit": "subreddit1",
                    "created_utc": 1712345678,
                    "num_comments": 10,
                    "score": 100,
                    "upvote_ratio": 0.5,
                    "upvotes": 50,
                    "permalink": "permalink1",
                    "url": "url1",
                    "is_video": False,
                },
            ],
        ),
        ({"data": {"after": None, "children": []}}, None, []),
    ],
)
async def test_parse_user_posts_response_without_body(posts_data, expected_cursor, expected_posts):
    dummy_context = object()
    result = await parse_user_posts_response(dummy_context, posts_data, include_body=False)

    if expected_cursor:
        expected = {"cursor": expected_cursor, "posts": expected_posts}
    else:
        expected = {"posts": expected_posts}

    assert result == expected



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_hubspot
from arcade_hubspot.tools import get_company_data_by_keywords

rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
catalog.add_module(arcade_hubspot)


@tool_eval()
def get_company_data_by_keywords_eval_suite() -> EvalSuite:
    """Create an evaluation suite for the get_company_data_by_keywords tool."""
    suite = EvalSuite(
        name="Get Company Data by Keywords",
        system_message="You are an AI assistant that can interact with Hubspot CRM using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get company data by keywords",
        user_message="Get information about the Acme company.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_company_data_by_keywords,
                args={
                    "keywords": "Acme",
                    "limit": 10,
                    "next_page_token": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.8),
            BinaryCritic(critic_field="next_page_token", weight=0.2),
        ],
    )

    suite.add_case(
        name="Get company data by keywords with limit",
        user_message="Get information of up to 3 companies with the word 'Acme' in their name.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_company_data_by_keywords,
                args={
                    "keywords": "Acme",
                    "limit": 3,
                    "next_page_token": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.6),
            BinaryCritic(critic_field="limit", weight=0.3),
            BinaryCritic(critic_field="next_page_token", weight=0.1),
        ],
    )

    suite.add_case(
        name="Get summary of company activity",
        user_message="Get a summary of the latest activities in the Acme company.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_company_data_by_keywords,
                args={
                    "keywords": "Acme",
                    "limit": 10,
                    "next_page_token": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.8),
            BinaryCritic(critic_field="next_page_token", weight=0.2),
        ],
    )

    suite.add_case(
        name="Interactions with contacts of an account",
        user_message="Get me the interactions with the contacts of the Acme company.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_company_data_by_keywords,
                args={
                    "keywords": "Acme",
                    "limit": 10,
                    "next_page_token": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.8),
            BinaryCritic(critic_field="next_page_token", weight=0.2),
        ],
    )

    suite.add_case(
        name="Emails or calls with contacts of an account",
        user_message="Were there any emails or calls with contacts of the Acme company this week?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_company_data_by_keywords,
                args={
                    "keywords": "Acme",
                    "limit": 10,
                    "next_page_token": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.8),
            BinaryCritic(critic_field="next_page_token", weight=0.2),
        ],
    )

    suite.add_case(
        name="Get company status",
        user_message="What's the status of the Acme company?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_company_data_by_keywords,
                args={
                    "keywords": "Acme",
                    "limit": 10,
                    "next_page_token": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.8),
            BinaryCritic(critic_field="next_page_token", weight=0.2),
        ],
    )

    suite.add_case(
        name="Get overdue tasks",
        user_message="Are there any tasks overdue for the Acme company?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_company_data_by_keywords,
                args={
                    "keywords": "Acme",
                    "limit": 10,
                    "next_page_token": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.8),
            BinaryCritic(critic_field="next_page_token", weight=0.2),
        ],
    )

    suite.add_case(
        name="Get company closing likelihood",
        user_message="What's the likelihood of the Acme company closing a deal?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_company_data_by_keywords,
                args={
                    "keywords": "Acme",
                    "limit": 10,
                    "next_page_token": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.8),
            BinaryCritic(critic_field="next_page_token", weight=0.2),
        ],
    )

    return suite



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Google

from arcade_google.utils import build_docs_service


# Uses https://developers.google.com/docs/api/reference/rest/v1/documents/get
# Example `arcade chat` query: `get document with ID 1234567890`
# Note: Document IDs are returned in the response of the Google Drive's `list_documents` tool
@tool(
    requires_auth=Google(
        scopes=[
            "https://www.googleapis.com/auth/drive.file",
        ],
    )
)
async def get_document_by_id(
    context: ToolContext,
    document_id: Annotated[str, "The ID of the document to retrieve."],
) -> Annotated[dict, "The document contents as a dictionary"]:
    """
    Get the latest version of the specified Google Docs document.
    """
    service = build_docs_service(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    # Execute the documents().get() method. Returns a Document object
    # https://developers.google.com/docs/api/reference/rest/v1/documents#Document
    request = service.documents().get(documentId=document_id)
    response = request.execute()
    return dict(response)


# Uses https://developers.google.com/docs/api/reference/rest/v1/documents/batchUpdate
# Example `arcade chat` query: `insert "The END" at the end of document with ID 1234567890`
@tool(
    requires_auth=Google(
        scopes=[
            "https://www.googleapis.com/auth/drive.file",
        ],
    )
)
async def insert_text_at_end_of_document(
    context: ToolContext,
    document_id: Annotated[str, "The ID of the document to update."],
    text_content: Annotated[str, "The text content to insert into the document"],
) -> Annotated[dict, "The response from the batchUpdate API as a dict."]:
    """
    Updates an existing Google Docs document using the batchUpdate API endpoint.
    """
    document = await get_document_by_id(context, document_id)

    end_index = document["body"]["content"][-1]["endIndex"]

    service = build_docs_service(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    requests = [
        {
            "insertText": {
                "location": {
                    "index": int(end_index) - 1,
                },
                "text": text_content,
            }
        }
    ]

    # Execute the documents().batchUpdate() method
    response = (
        service.documents()
        .batchUpdate(documentId=document_id, body={"requests": requests})
        .execute()
    )

    return dict(response)


# Uses https://developers.google.com/docs/api/reference/rest/v1/documents/create
# Example `arcade chat` query: `create blank document with title "My New Document"`
@tool(
    requires_auth=Google(
        scopes=[
            "https://www.googleapis.com/auth/drive.file",
        ],
    )
)
async def create_blank_document(
    context: ToolContext, title: Annotated[str, "The title of the blank document to create"]
) -> Annotated[dict, "The created document's title, documentId, and documentUrl in a dictionary"]:
    """
    Create a blank Google Docs document with the specified title.
    """
    service = build_docs_service(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    body = {"title": title}

    # Execute the documents().create() method. Returns a Document object https://developers.google.com/docs/api/reference/rest/v1/documents#Document
    request = service.documents().create(body=body)
    response = request.execute()

    return {
        "title": response["title"],
        "documentId": response["documentId"],
        "documentUrl": f"https://docs.google.com/document/d/{response['documentId']}/edit",
    }


# Uses https://developers.google.com/docs/api/reference/rest/v1/documents/batchUpdate
# Example `arcade chat` query:
#   `create document with title "My New Document" and text content "Hello, World!"`
@tool(
    requires_auth=Google(
        scopes=[
            "https://www.googleapis.com/auth/drive.file",
        ],
    )
)
async def create_document_from_text(
    context: ToolContext,
    title: Annotated[str, "The title of the document to create"],
    text_content: Annotated[str, "The text content to insert into the document"],
) -> Annotated[dict, "The created document's title, documentId, and documentUrl in a dictionary"]:
    """
    Create a Google Docs document with the specified title and text content.
    """
    # First, create a blank document
    document = await create_blank_document(context, title)

    service = build_docs_service(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    requests = [
        {
            "insertText": {
                "location": {
                    "index": 1,
                },
                "text": text_content,
            }
        }
    ]

    # Execute the batchUpdate method to insert text
    service.documents().batchUpdate(
        documentId=document["documentId"], body={"requests": requests}
    ).execute()

    return {
        "title": document["title"],
        "documentId": document["documentId"],
        "documentUrl": f"https://docs.google.com/document/d/{document['documentId']}/edit",
    }



from typing import Annotated

from arcade_tdk import ToolContext, tool
from e2b_code_interpreter import Sandbox

from arcade_code_sandbox.tools.models import E2BSupportedLanguage

# See https://e2b.dev/docs to learn more about E2B


@tool(requires_secrets=["E2B_API_KEY"])
def run_code(
    context: ToolContext,
    code: Annotated[str, "The code to run"],
    language: Annotated[
        E2BSupportedLanguage, "The language of the code"
    ] = E2BSupportedLanguage.PYTHON,
) -> Annotated[str, "The sandbox execution as a JSON string"]:
    """
    Run code in a sandbox and return the output.
    """
    api_key = context.get_secret("E2B_API_KEY")

    with Sandbox(api_key=api_key) as sbx:
        execution = sbx.run_code(code=code, language=language)

    return str(execution.to_json())


# Note: Not recommended to use tool_choice='generate' with this tool
#       since it contains base64 encoded image.
@tool(requires_secrets=["E2B_API_KEY"])
def create_static_matplotlib_chart(
    context: ToolContext,
    code: Annotated[str, "The Python code to run"],
) -> Annotated[dict, "A dictionary with the following keys: base64_image, logs, error"]:
    """
    Run the provided Python code to generate a static matplotlib chart.
    The resulting chart is returned as a base64 encoded image.
    """
    api_key = context.get_secret("E2B_API_KEY")

    with Sandbox(api_key=api_key) as sbx:
        execution = sbx.run_code(code=code)

    result = {
        "base64_image": execution.results[0].png if execution.results else None,
        "logs": execution.logs.to_json(),
        "error": execution.error.to_json() if execution.error else None,
    }

    return result



from typing import Annotated

from arcade_tdk import tool


@tool
def say_hello(name: Annotated[str, "The name of the person to greet"]) -> str:
    """Say a greeting!"""

    return "Hello, " + name + "!"



from unittest.mock import MagicMock, patch

import pytest
from arcade_core.catalog import ToolCatalog
from arcade_core.errors import ToolDefinitionError
from arcade_core.schema import FullyQualifiedName
from arcade_core.toolkit import Toolkit
from arcade_tdk import tool


@tool
def sample_tool() -> str:
    """
    A sample tool function
    """
    return "Hello, world!"


def test_add_tool_with_empty_toolkit_name_raises():
    catalog = ToolCatalog()
    with pytest.raises(ValueError):
        catalog.add_tool(sample_tool, "")


def test_add_tool_with_toolkit_name():
    catalog = ToolCatalog()
    catalog.add_tool(sample_tool, "sample_toolkit")
    assert (
        catalog.get_tool(FullyQualifiedName("SampleTool", "SampleToolkit", None)).tool
        == sample_tool
    )


def test_add_tool_with_toolkit():
    catalog = ToolCatalog()
    toolkit = Toolkit(
        name="sample_toolkit",
        description="A sample toolkit",
        version="1.0.0",
        package_name="sample_toolkit",
    )
    catalog.add_tool(sample_tool, toolkit)
    assert (
        catalog.get_tool(FullyQualifiedName("SampleTool", "SampleToolkit", None)).tool
        == sample_tool
    )


@pytest.mark.parametrize(
    "toolkit_version, expected_tool",
    [
        ("1.0.0", sample_tool),
        (None, sample_tool),
    ],
)
def test_get_tool(toolkit_version: str | None, expected_tool):
    catalog = ToolCatalog()
    fake_toolkit = Toolkit(
        name="sample_toolkit",
        description="A sample toolkit",
        version="1.0.0",
        package_name="sample_toolkit",
    )
    catalog.add_tool(sample_tool, fake_toolkit, module=None)

    fq_name = FullyQualifiedName(
        name="SampleTool", toolkit_name="SampleToolkit", toolkit_version=toolkit_version
    )
    tool = catalog.get_tool(fq_name)
    assert tool.tool == expected_tool


def test_add_toolkit_type_error():
    catalog = ToolCatalog()

    # Create a mock toolkit with an invalid tool
    class InvalidTool:
        pass

    mock_toolkit = Toolkit(
        name="mock_toolkit",
        description="A mock toolkit",
        version="0.0.1",
        package_name="mock_toolkit",
    )
    mock_toolkit.tools = {"mock_module": ["invalid_tool"]}

    # Mock the import_module and getattr functions
    with (
        patch("arcade_core.catalog.import_module") as mock_import,
        patch("arcade_core.catalog.getattr") as mock_getattr,
    ):
        mock_import.return_value = MagicMock()
        mock_getattr.return_value = InvalidTool()

        # Assert that ToolDefinitionError is raised
        with pytest.raises(ToolDefinitionError):
            catalog.add_toolkit(mock_toolkit)


def test_get_tool_by_name():
    catalog = ToolCatalog()
    catalog.add_tool(sample_tool, "sample_toolkit")

    tool = catalog.get_tool_by_name("SampleToolkit.SampleTool")
    assert tool.tool == sample_tool
    assert tool.name == "SampleTool"
    assert tool.meta.toolkit == "sample_toolkit"
    assert tool.version is None

    with pytest.raises(ValueError):
        catalog.get_tool_by_name("nonexistent_toolkit.SampleTool")


def test_get_tool_by_name_with_version():
    catalog = ToolCatalog()
    catalog.add_tool(sample_tool, "sample_toolkit")

    tool = catalog.get_tool_by_name("SampleToolkit.SampleTool")
    assert tool.tool == sample_tool
    assert tool.name == "SampleTool"
    assert tool.meta.toolkit == "sample_toolkit"

    with pytest.raises(ValueError):
        catalog.get_tool_by_name("SampleToolkit.SampleTool", version="2.0.0")


def test_get_tool_by_name_with_invalid_version():
    catalog = ToolCatalog()
    catalog.add_tool(sample_tool, "SampleToolkit")

    with pytest.raises(ValueError):
        catalog.get_tool_by_name("SampleToolkit.SampleTool", version="2.0.0")


def test_load_disabled_tools(monkeypatch):
    disabled_tools = (
        "SampleToolkitOne.SampleToolOne,"  # valid
        + "SampleToolkitOne_SampleToolTwo,"  # invalid
        + "SampleToolkitTwo.SampleToolThree,"  # valid
        + "SampleToolkitTwo.SampleToolFour@0.0.1,"  # invalid
        + "SampleToolkitThree_SampleToolFive@0.0.1,"  # invalid
        + "SampleToolkitFour.sample_tool_six,"  # invalid
        + "sample_toolkit5.SampleTool7,"  # invalid
        + "sample_toolkit6.sample_tool_8"  # invalid
    )
    expected_disabled_tools = {
        "sampletoolkitone.sampletoolone",
        "sampletoolkittwo.sampletoolthree",
    }

    monkeypatch.setenv("ARCADE_DISABLED_TOOLS", disabled_tools)
    catalog = ToolCatalog()

    assert catalog._disabled_tools == expected_disabled_tools


def test_add_tool_with_disabled_tool(monkeypatch):
    monkeypatch.setenv("ARCADE_DISABLED_TOOLS", "SampleToolkitOne.SampleTool")
    catalog = ToolCatalog()

    catalog.add_tool(sample_tool, "SampleToolkitOne")
    assert len(catalog._tools) == 0


def test_add_tool_with_empty_string_disabled_tools(monkeypatch):
    monkeypatch.setenv("ARCADE_DISABLED_TOOLS", "")
    catalog = ToolCatalog()
    catalog.add_tool(sample_tool, "SampleToolkitOne")
    assert len(catalog._tools) == 1


def test_add_tool_with_whitespace_disabled_tools(monkeypatch):
    monkeypatch.setenv("ARCADE_DISABLED_TOOLS", "            ")
    catalog = ToolCatalog()
    catalog.add_tool(sample_tool, "SampleToolkitOne")
    assert len(catalog._tools) == 1


def test_add_tool_with_disabled_toolkit(monkeypatch):
    monkeypatch.setenv("ARCADE_DISABLED_TOOLKITS", "SampleToolkitOne")
    catalog = ToolCatalog()

    catalog.add_toolkit(
        Toolkit(
            name="SampleToolkitOne",
            package_name="sample_toolkit_one",
            version="1.0.0",
            description="A sample toolkit",
        )
    )
    assert len(catalog._tools) == 0



import os
from datetime import datetime

from configuration import AgentConfigurable
from langchain_arcade import ToolManager
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI
from langgraph.errors import NodeInterrupt
from langgraph.graph import END, START, MessagesState, StateGraph
from langgraph.prebuilt import ToolNode

# Initialize the Arcade Tool Manager with your API key
arcade_api_key = os.getenv("ARCADE_API_KEY")
openai_api_key = os.getenv("OPENAI_API_KEY")

manager = ToolManager(api_key=arcade_api_key)
manager.init_tools(toolkits=["Github"])
tools = manager.to_langchain(use_interrupts=True)
tool_node = ToolNode(tools)

PROMPT_TEMPLATE = f"""
You are a helpful assistant who can use tools to help users with tasks
Today's date is {datetime.now().strftime("%Y-%m-%d")}

ALL RESPONSES should be in plain text and not markdown.
"""
# prompt for the main agent
prompt = ChatPromptTemplate.from_messages([
    ("system", PROMPT_TEMPLATE),
    ("placeholder", "{messages}"),
])
# Initialize the language model with your OpenAI API key
model = ChatOpenAI(model="gpt-4o", api_key=openai_api_key).bind_tools(tools)
prompted_model = prompt | model


class AgentState(MessagesState):
    auth_url: str | None = None


def call_agent(state):
    """Define the agent function that invokes the model"""
    messages = state["messages"]
    # replace placeholder with messages from state
    response = prompted_model.invoke({"messages": messages})
    return {"messages": [response]}


def should_continue(state: AgentState, config: dict):
    """Function to determine the next step based on the model's response"""
    last_message = state["messages"][-1]
    if last_message.tool_calls:
        return "check_auth"
    # If no tool calls are present, end the workflow
    return END


def check_auth(state: AgentState, config: dict):
    user_id = config["configurable"].get("user_id")
    tool_name = state["messages"][-1].tool_calls[0]["name"]
    auth_response = manager.authorize(tool_name, user_id)
    if auth_response.status != "completed":
        return {"auth_url": auth_response.url}
    else:
        return {"auth_url": None}


def authorize(state: AgentState, config: dict):
    """Function to handle tool authorization"""
    user_id = config["configurable"].get("user_id")
    tool_name = state["messages"][-1].tool_calls[0]["name"]
    auth_response = manager.authorize(tool_name, user_id)
    if auth_response.status != "completed":
        auth_message = (
            f"Please authorize the application in your browser:\n\n {state.get('auth_url')}"
        )
        raise NodeInterrupt(auth_message)


# Build the workflow graph
workflow = StateGraph(AgentState, AgentConfigurable)

# Add nodes to the graph
workflow.add_node("agent", call_agent)
workflow.add_node("tools", tool_node)
workflow.add_node("authorization", authorize)
workflow.add_node("check_auth", check_auth)

# Define the edges and control flow
workflow.add_edge(START, "agent")
workflow.add_conditional_edges("agent", should_continue, ["check_auth", END])
workflow.add_edge("check_auth", "authorization")
workflow.add_edge("authorization", "tools")
workflow.add_edge("tools", "agent")

# Compile the graph with an interrupt after the authorization node
# so that we can prompt the user to authorize the application
graph = workflow.compile()



from arcade_google_docs.tools import (
    create_blank_document,
    create_document_from_text,
    get_document_by_id,
    insert_text_at_end_of_document,
    search_and_retrieve_documents,
    search_documents,
)

__all__ = [
    "create_blank_document",
    "create_document_from_text",
    "get_document_by_id",
    "insert_text_at_end_of_document",
    "search_and_retrieve_documents",
    "search_documents",
]



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_google_drive
from arcade_google_drive.tools import (
    get_file_tree_structure,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)

catalog = ToolCatalog()
catalog.add_module(arcade_google_drive)


@tool_eval()
def get_file_tree_structure_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Google Drive tools."""
    suite = EvalSuite(
        name="Google Drive Tools Evaluation",
        system_message="You are an AI assistant that can manage Google Drive documents using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="get my google drive's file tree structure including shared drives",
        user_message="get my google drive's file tree structure including shared drives",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_file_tree_structure,
                args={
                    "restrict_to_shared_drive_id": None,
                    "include_shared_drives": True,
                    "include_organization_domain_documents": False,
                    "order_by": None,
                    "limit": None,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_shared_drives", weight=0.5),
            BinaryCritic(critic_field="restrict_to_shared_drive_id", weight=0.5 / 4),
            BinaryCritic(critic_field="include_organization_domain_documents", weight=0.5 / 4),
            BinaryCritic(critic_field="order_by", weight=0.5 / 4),
            BinaryCritic(critic_field="limit", weight=0.5 / 4),
        ],
    )

    suite.add_case(
        name="get my google drive's file tree structure without shared drives",
        user_message="get my google drive's file tree structure without shared drives",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_file_tree_structure,
                args={
                    "restrict_to_shared_drive_id": None,
                    "include_shared_drives": False,
                    "include_organization_domain_documents": False,
                    "order_by": None,
                    "limit": None,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_shared_drives", weight=0.5),
            BinaryCritic(critic_field="restrict_to_shared_drive_id", weight=0.5 / 4),
            BinaryCritic(critic_field="include_organization_domain_documents", weight=0.5 / 4),
            BinaryCritic(critic_field="order_by", weight=0.5 / 4),
            BinaryCritic(critic_field="limit", weight=0.5 / 4),
        ],
    )

    suite.add_case(
        name="what are the files in the folder 'hello world' in my google drive?",
        user_message="what are the files in the folder 'hello world' in my google drive?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_file_tree_structure,
                args={
                    "restrict_to_shared_drive_id": None,
                    "include_shared_drives": False,
                    "include_organization_domain_documents": False,
                    "order_by": None,
                    "limit": None,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_shared_drives", weight=0.5),
            BinaryCritic(critic_field="restrict_to_shared_drive_id", weight=0.5 / 4),
            BinaryCritic(critic_field="include_organization_domain_documents", weight=0.5 / 4),
            BinaryCritic(critic_field="order_by", weight=0.5 / 4),
            BinaryCritic(critic_field="limit", weight=0.5 / 4),
        ],
    )

    suite.add_case(
        name="how many files are there in all my google drives, including shared ones?",
        user_message="how many files are there in all my google drives, including shared ones?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_file_tree_structure,
                args={
                    "restrict_to_shared_drive_id": None,
                    "include_shared_drives": True,
                    "include_organization_domain_documents": False,
                    "order_by": None,
                    "limit": None,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_shared_drives", weight=0.5),
            BinaryCritic(critic_field="restrict_to_shared_drive_id", weight=0.5 / 4),
            BinaryCritic(critic_field="include_organization_domain_documents", weight=0.5 / 4),
            BinaryCritic(critic_field="order_by", weight=0.5 / 4),
            BinaryCritic(critic_field="limit", weight=0.5 / 4),
        ],
    )

    return suite



import pytest
from msgraph.generated.models.email_address import EmailAddress as GraphEmailAddress
from msgraph.generated.models.recipient import Recipient as GraphRecipient

from arcade_outlook_mail.message import Recipient


@pytest.mark.parametrize(
    "input_sdk_recipient, expected_email, expected_name",
    [
        (
            GraphRecipient(email_address=GraphEmailAddress(address="dev@arcade.dev", name="Dev")),
            "dev@arcade.dev",
            "Dev",
        ),
        (
            GraphRecipient(email_address=GraphEmailAddress(address="dev@arcade.dev")),
            "dev@arcade.dev",
            "",
        ),
        (GraphRecipient(email_address=GraphEmailAddress(name="Dev")), "", "Dev"),
        (GraphRecipient(email_address=GraphEmailAddress()), "", ""),
        (GraphRecipient(), "", ""),
    ],
)
def test_recipient(input_sdk_recipient, expected_email, expected_name):
    recipient = Recipient.from_sdk(input_sdk_recipient)
    assert (
        recipient.email_address == expected_email
    ), "SDK conversion didn't set email_address correctly"
    assert recipient.name == expected_name, "SDK conversion didn't set name correctly"

    recipient_dict = recipient.to_dict()
    expected_dict = {"email_address": expected_email, "name": expected_name}
    assert recipient_dict == expected_dict, "to_dict conversion did not produce expected dictionary"

    actual_sdk_recipient = recipient.to_sdk()
    assert (
        actual_sdk_recipient.email_address.address == expected_email
    ), "to_sdk conversion produced wrong email address"
    assert (
        actual_sdk_recipient.email_address.name == expected_name
    ), "to_sdk conversion produced wrong name"



from arcade_tdk.errors import ToolExecutionError


class AsanaToolExecutionError(ToolExecutionError):
    pass


class PaginationTimeoutError(AsanaToolExecutionError):
    def __init__(self, timeout_seconds: int, tool_name: str):
        message = f"Pagination timed out after {timeout_seconds} seconds"
        super().__init__(
            message=message,
            developer_message=f"{message} while calling the tool {tool_name}",
        )



import json

import httpx
import pytest

from arcade_jira.client import JiraClient
from arcade_jira.exceptions import JiraToolExecutionError


@pytest.mark.asyncio
async def test_get_cloud_data_from_available_resources_single_cloud(
    mock_httpx_client, fake_auth_token
):
    cloud = {"id": "123", "name": "Test Cloud", "url": "https://test.atlassian.net"}

    client = JiraClient(auth_token=fake_auth_token)

    mock_httpx_client.get.return_value = httpx.Response(
        status_code=200,
        json=[cloud],
    )

    response = await client._get_cloud_data_from_available_resources()
    assert response == cloud


@pytest.mark.asyncio
async def test_get_cloud_data_from_available_resources_multiple_clouds(
    mock_httpx_client, fake_auth_token
):
    cloud1 = {"id": "123", "name": "Test Cloud", "url": "https://test.atlassian.net"}
    cloud2 = {"id": "456", "name": "Test Cloud 2", "url": "https://test2.atlassian.net"}

    client = JiraClient(auth_token=fake_auth_token)

    mock_httpx_client.get.return_value = httpx.Response(
        status_code=200,
        json=[cloud1, cloud2],
    )

    with pytest.raises(JiraToolExecutionError) as error:
        await client._get_cloud_data_from_available_resources()

    assert "Multiple cloud IDs returned by Atlassian" in error.value.message
    assert json.dumps(cloud1) in error.value.message
    assert json.dumps(cloud2) in error.value.message


@pytest.mark.asyncio
async def test_get_cloud_data_from_available_resources_duplicate_cloud(
    mock_httpx_client, fake_auth_token
):
    cloud = {"id": "123", "name": "Test Cloud", "url": "https://test.atlassian.net"}

    client = JiraClient(auth_token=fake_auth_token)

    mock_httpx_client.get.return_value = httpx.Response(
        status_code=200,
        json=[cloud, cloud],
    )

    response = await client._get_cloud_data_from_available_resources()
    assert response == cloud



update_draft_email_additional_messages = [
    {"role": "system", "content": "Today is 2025-04-22, Tuesday."},
    {
        "role": "user",
        "content": '"create a new draft email with subject \'Hello friends\' and body "\n"\'I\'ve gathered you all here to celebrate the launch of the new Arcade platform."\n            "address it to e@arcade.dev and z@arcade.dev. also carbon copy to j@arcade.dev, "\n"f@arcade.dev, k@arcade.dev and finally to m@arcade.dev. also bcc to r@arcade.dev"',
    },
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_lKw4S01FGe03oZeuW25roepy",
                "type": "function",
                "function": {
                    "name": "Microsoft_CreateDraftEmail",
                    "arguments": '{"subject":"Hello friends","body":"I\'ve gathered you all here to celebrate the launch of the new Arcade platform.","to_recipients":["e@arcade.dev","z@arcade.dev"],"cc_recipients":["j@arcade.dev","f@arcade.dev","k@arcade.dev","m@arcade.dev"],"bcc_recipients":["r@arcade.dev"]}',
                },
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"bcc_recipients":[{"email_address":"r@arcade.dev","name":"r@arcade.dev"}],"body":"I\'ve gathered you all here to celebrate the launch of the new Arcade platform.","cc_recipients":[{"email_address":"j@arcade.dev","name":"j@arcade.dev"},{"email_address":"f@arcade.dev","name":"f@arcade.dev"},{"email_address":"k@arcade.dev","name":"k@arcade.dev"},{"email_address":"m@arcade.dev","name":"m@arcade.dev"}],"conversation_id":"AQQkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoAEAAskq2oM-moTbt3gDT_yK0e","conversation_index":"AQHbs6c0LJKtqDP5qE27d4A0/sitHg==","flag":{"due_date_time":"","flag_status":"notFlagged"},"from":{"email_address":"","name":""},"has_attachments":false,"importance":"normal","is_draft":true,"is_read":true,"message_id":"AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4_x_WeUwAAAIBDwAAAFuxokOLZRtDncM4_x_WeUwAAAAC-dpvAAAA","received_date_time":"2025-04-22T16:54:25+00:00","reply_to":[],"subject":"Hello friends","to_recipients":[{"email_address":"e@arcade.dev","name":"e@arcade.dev"},{"email_address":"z@arcade.dev","name":"z@arcade.dev"}],"web_link":"https://outlook.live.com/owa/?ItemID=AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4%2Bx%2BWeUwAAAIBDwAAAFuxokOLZRtDncM4%2Bx%2BWeUwAAAAC%2FdpvAAAA\\u0026exvsurl=1\\u0026viewmodel=ReadMessageItem"}',
        "tool_call_id": "call_lKw4S01FGe03oZeuW25roepy",
        "name": "Microsoft_CreateDraftEmail",
    },
    {
        "role": "assistant",
        "content": 'I have created a draft email with the subject "Hello friends" addressed to the specified recipients. You can view and edit the draft through [this link](https://outlook.live.com/owa/?ItemID=AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4%2Bx%2BWeUwAAAIBDwAAAFuxokOLZRtDncM4%2Bx%2BWeUwAAAAC%2FdpvAAAA&exvsurl=1&viewmodel=ReadMessageItem).',
    },
]

list_emails_with_pagination_token_additional_messages = [
    {"role": "system", "content": "Today is 2025-04-21, Monday."},
    {"role": "user", "content": "get one email"},
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_jACvc3Gl1WHkqWgI8gdsIt0G",
                "type": "function",
                "function": {"name": "Microsoft_ListEmails", "arguments": '{"limit":1}'},
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"messages":[{"bcc_recipients":[],"body":"Microsoft account New app(s) have access to your data Arcade","cc_recipients":["e@arcade.dev],"conversation_id":"AQQkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoAEABOD15A17tWSaVHkmjhko1R","conversation_index":"AQHbsuDd Tg9eQNe7VkmlR5Jo4ZKNUQ==","flag":{"due_date_time":"","flag_status":"notFlagged"},"from":{"email_address":"account-security@accountprotect ion.microsoft.com","name":"Microsoft account team"},"has_attachments":false,"importance":"normal","is_draft":false,"is_read":true,"message_id":"AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4_x_WeUwAAAIBDAAAAFuxokOLZRtDncM4_x_WeUwAAAABc_ezAAAA","received_date_time":"2025-04-21T17:14:39+00:00", "reply_to":[],"subject":"New app(s) connected to your Microsoft account","to_recipients":[{"email_address":"ericarcade@outlook.com","name":"ericarcade@outlook.com"}],"web_link":"https://outlook.live.com/owa/?ItemID=AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4%2Bx%2BWeUwAAAIBDAAAAFuxokOLZRtDncM4%2Bx%2BWeUwAAAABc%2BezAAAA\\u0026exvsurl=1\\u0026viewmodel=ReadMessageItem"}],"num_messages":1,"pagination_token":"https://graph.microsoft.com/v1.0/me/messages?%24count=true&%24orderby=receivedDateTime+DESC&%24select=bccRecipients%2cbody%2cccRecipients%2cconversationId%2cconversationIndex%2cflag%2cfrom%2chasAttachments%2cimportance%2cisDraft%2cisRead%2creceivedDateTime%2creplyTo%2csubject%2ctoRecipients%2cwebLink&%24top=1&%24skip=1"}',
        "tool_call_id": "call_jACvc3Gl1WHkqWgI8gdsIt0G",
        "name": "Microsoft_ListEmails",
    },
    {
        "role": "assistant",
        "content": "Here is the most recent email you received:\n\n- **From:** Microsoft account team (account-security@accountprotection.microsoft.com)\n- **To:** e@outlook.com\n- **Subject:** New app(s) connected to your Microsoft account\n- **Received Date:** April 21, 2025\n- **Body:**\n  ```\n  Microsoft account\n\n  New app(s) have access to your data Arcade connected to the Microsoft account *@outlook.com.```\n- **Link to email:** [Read in Outlook](https://outlook.live.com/owa/?ItemID=AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4%2Bx%2BWeUwAAAIBDAAAAFuxokOLZRtDncM4%2Bx%2BWeUwAAAABc%2BezAAAA&exvsurl=1&viewmodel=ReadMessageItem)",
    },
]

list_emails_with_pagination_token_additional_messages = [
    {"role": "system", "content": "Today is 2025-04-21, Monday."},
    {"role": "user", "content": "get one email"},
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_jACvc3Gl1WHkqWgI8gdsIt0G",
                "type": "function",
                "function": {"name": "Microsoft_ListEmails", "arguments": '{"limit":1}'},
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"messages":[{"bcc_recipients":[],"body":"Microsoft account New app(s) have access to your data Arcade","cc_recipients":[],"conversation_id":"AQQkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoAEABOD15A17tWSaVHkmjhko1R","conversation_index":"AQHbsuDdTg9eQNe7VkmlR5Jo4ZKNUQ==","flag":{"due_date_time":"","flag_status":"notFlagged"},"from":{"email_address":"account-security-noreply@accountprotection.microsoft.com","name":"Microsoft account team"},"has_attachments":false,"importance":"normal","is_draft":false,"is_read":true,"message_id":"AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4_x_WeUwAAAIBDAAAAFuxokOLZRtDncM4_x_WeUwAAAABc_ezAAAA","received_date_time":"2025-04-21T17:14:39+00:00", "reply_to":[],"subject":"New app(s) connected to your Microsoft account","to_recipients":[{"email_address":"ericarcade@outlook.com","name":"ericarcade@outlook.com"}],"web_link":"https://outlook.live.com/owa/?I temID=AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4%2Bx%2BWeUwAAAIBDAAAAFuxokOLZRtDncM4%2Bx%2BWeUwAAAA Bc%2BezAAAA\\u0026exvsurl=1\\u0026viewmodel=ReadMessageItem"}],"num_messages":1,"pagination_token":"https://graph.microsoft.com/v1.0/me/messages?%24count=true&%24orderby=receivedDateTime+DESC&%24select=bccRecipients%2cbody%2cccRecipients%2cconversationId%2cconversationIndex%2cflag%2cfrom%2chasAttachments%2cimportance%2cisDraft%2cisRead%2creceivedDateTime%2creplyTo%2csubject%2ctoRecipients%2cwebLink&%24top=1&%24skip=1"}',
        "tool_call_id": "call_jACvc3Gl1WHkqWgI8gdsIt0G",
        "name": "Microsoft_ListEmails",
    },
    {
        "role": "assistant",
        "content": "Here is the most recent email you received:\n\n- **From:** Microsoft account team (account-security-noreply@accountprotection.microsoft.com)\n- **To:** e@outlook.com\n- **Subject:** New app(s) connected to your Microsoft account\n- **Received Date:** April 21, 2025\n- **Body:**\n  ```\n  Microsoft account\n\n  New app(s) have access to your data Arcade connected to the Microsoft account *@outlook.com.```\n- **Link to email:** [Read in Outlook](https://outlook.live.com/owa/?ItemID=AQMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMHAFuxokOLZRtDncM4%2Bx%2BWeUwAAAIBDAAAAFuxokOLZRtDncM4%2Bx%2BWeUwAAAABc%2BezAAAA&exvsurl=1&viewmodel=ReadMessageItem)",
    },
]



from typing import Annotated, Any

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Asana

from arcade_asana.constants import TEAM_OPT_FIELDS
from arcade_asana.models import AsanaClient
from arcade_asana.utils import (
    get_next_page,
    get_unique_workspace_id_or_raise_error,
    remove_none_values,
)


@tool(requires_auth=Asana(scopes=["default"]))
async def get_team_by_id(
    context: ToolContext,
    team_id: Annotated[str, "The ID of the Asana team to get"],
) -> Annotated[dict[str, Any], "Get an Asana team by its ID"]:
    """Get an Asana team by its ID"""
    client = AsanaClient(context.get_auth_token_or_empty())
    response = await client.get(
        f"/teams/{team_id}",
        params=remove_none_values({"opt_fields": TEAM_OPT_FIELDS}),
    )
    return {"team": response["data"]}


@tool(requires_auth=Asana(scopes=["default"]))
async def list_teams_the_current_user_is_a_member_of(
    context: ToolContext,
    workspace_id: Annotated[
        str | None,
        "The workspace ID to list teams from. Defaults to None. If no workspace ID is provided, "
        "it will use the current user's workspace , if there's only one. If the user has multiple "
        "workspaces, it will raise an error.",
    ] = None,
    limit: Annotated[
        int, "The maximum number of teams to return. Min is 1, max is 100. Defaults to 100."
    ] = 100,
    next_page_token: Annotated[
        str | None,
        "The token to retrieve the next page of teams. Defaults to None (start from the first page "
        "of teams)",
    ] = None,
) -> Annotated[
    dict[str, Any],
    "List teams in Asana that the current user is a member of",
]:
    """List teams in Asana that the current user is a member of"""
    limit = max(1, min(100, limit))

    workspace_id = workspace_id or await get_unique_workspace_id_or_raise_error(context)

    client = AsanaClient(context.get_auth_token_or_empty())
    response = await client.get(
        "/users/me/teams",
        params=remove_none_values({
            "limit": limit,
            "offset": next_page_token,
            "opt_fields": TEAM_OPT_FIELDS,
            "organization": workspace_id,
        }),
    )

    return {
        "teams": response["data"],
        "count": len(response["data"]),
        "next_page": get_next_page(response),
    }


@tool(requires_auth=Asana(scopes=["default"]))
async def list_teams(
    context: ToolContext,
    workspace_id: Annotated[
        str | None,
        "The workspace ID to list teams from. Defaults to None. If no workspace ID is provided, "
        "it will use the current user's workspace, if there's only one. If the user has multiple "
        "workspaces, it will raise an error listing the available workspaces.",
    ] = None,
    limit: Annotated[
        int, "The maximum number of teams to return. Min is 1, max is 100. Defaults to 100."
    ] = 100,
    next_page_token: Annotated[
        str | None,
        "The token to retrieve the next page of teams. Defaults to None (start from the first page "
        "of teams)",
    ] = None,
) -> Annotated[dict[str, Any], "List teams in an Asana workspace"]:
    """List teams in an Asana workspace"""
    limit = max(1, min(100, limit))

    workspace_id = workspace_id or await get_unique_workspace_id_or_raise_error(context)

    client = AsanaClient(context.get_auth_token_or_empty())
    response = await client.get(
        f"/workspaces/{workspace_id}/teams",
        params=remove_none_values({
            "limit": limit,
            "offset": next_page_token,
            "opt_fields": TEAM_OPT_FIELDS,
        }),
    )

    return {
        "teams": response["data"],
        "count": len(response["data"]),
        "next_page": get_next_page(response),
    }



from unittest.mock import patch

import pytest
from arcade_tdk import ToolAuthorizationContext, ToolContext


@pytest.fixture
def mock_context():
    mock_auth = ToolAuthorizationContext(token="fake-token")  # noqa: S106
    return ToolContext(authorization=mock_auth)


@pytest.fixture
def mock_httpx_client():
    with patch("arcade_asana.models.httpx") as mock_httpx:
        yield mock_httpx.AsyncClient().__aenter__.return_value



from typing import Annotated, Any

import httpx
from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Notion
from arcade_tdk.errors import ToolExecutionError

from arcade_notion_toolkit.enums import ObjectType, SortDirection
from arcade_notion_toolkit.utils import (
    build_workspace_structure,
    get_headers,
    get_url,
    remove_none_values,
    simplify_search_result,
)


@tool(requires_auth=Notion())
async def search_by_title(
    context: ToolContext,
    query: Annotated[
        str | None,
        "A substring to search for within page and database titles. "
        "If not provided (default), all pages and/or databases are returned.",
    ] = None,
    select: Annotated[
        ObjectType | None,
        "Limit the results to either only pages or only databases. Defaults to both.",
    ] = None,
    order_by: Annotated[
        SortDirection,
        "The direction to sort search results by last edited time. Defaults to 'descending'.",
    ] = SortDirection.DESCENDING,
    limit: Annotated[
        int,
        "The maximum number of results to return. Defaults to 100. Set to -1 for no limit.",
    ] = 100,
) -> Annotated[
    dict,
    "A dictionary containing minimal information about the pages and/or databases that have "
    "titles that are the best match for the query. Does not include content or location.",
]:
    """Search for similar titles of pages, databases, or both within the user's workspace.
    Does not include content.
    """
    results = []
    current_cursor = None

    url = get_url("search_by_title")
    headers = get_headers(context)
    payload = {
        "query": query,
        "page_size": 100 if limit == -1 else min(100, limit),
        "filter": {"property": "object", "value": select.value} if select else None,
        "sort": {"direction": order_by, "timestamp": "last_edited_time"},
    }
    payload = remove_none_values(payload)

    async with httpx.AsyncClient() as client:
        while True:
            if current_cursor:
                payload["start_cursor"] = current_cursor
            elif "start_cursor" in payload:
                del payload["start_cursor"]

            response = await client.post(url, headers=headers, json=payload)
            response.raise_for_status()
            data = response.json()

            page_results = [simplify_search_result(item) for item in data.get("results", [])]
            results.extend(page_results)

            # If a limit is set and we've reached or exceeded it, truncate the results.
            if limit is not None and len(results) >= limit:
                results = results[:limit]
                break

            if not data.get("has_more", False):
                break

            current_cursor = data.get("next_cursor")

    return {"results": results}


@tool(requires_auth=Notion())
async def get_object_metadata(
    context: ToolContext,
    object_title: Annotated[
        str | None, "Title of the page or database whose metadata to get"
    ] = None,
    object_id: Annotated[str | None, "ID of the page or database whose metadata to get"] = None,
    object_type: Annotated[
        ObjectType | None,
        "The type of object to match title to. Only used if `object_title` is provided. "
        "Defaults to both",
    ] = None,
) -> Annotated[dict[str, Any], "The metadata of the object"]:
    """Get the metadata of a Notion object (page or database) from its title or ID.

    One of `object_title` or `object_id` MUST be provided, but both cannot be provided.
    The title is case-insensitive and outer whitespace is ignored.

    An object's metadata includes it's id, various timestamps, properties, url, and more.
    """

    async def get_metadata_by_title(object_title: str) -> dict[str, Any]:
        candidates_response = await search_by_title(
            context,
            object_title,
            select=object_type,
            order_by=SortDirection.DESCENDING,
            limit=3,
        )

        if object_type:
            candidates: list[dict[str, Any]] = [
                page
                for page in candidates_response["results"]
                if page["object"] == object_type.value
            ]
        else:
            candidates = candidates_response["results"]

        normalized_title = object_title.lower().strip()
        error_msg = (
            f"The {object_type.value if object_type else 'object'} with "
            f"the title '{object_title}' could not be found. "
            "Either it does not exist, or it has not been shared with the integration."
        )

        if not candidates:
            raise ToolExecutionError(message=error_msg)

        for object_ in candidates:
            if object_["title"].lower().strip() == normalized_title:
                # object_ is either a page object: https://developers.notion.com/reference/page
                # or a database object: https://developers.notion.com/reference/database
                return object_

        raise ToolExecutionError(
            message=error_msg,
            developer_message=f"The closest matches are: {candidates}",
        )

    async def get_metadata_by_id(object_id: str) -> dict[str, Any]:
        url = get_url("retrieve_a_page", page_id=object_id)
        headers = get_headers(context)
        async with httpx.AsyncClient(timeout=10) as client:
            response = await client.get(url, headers=headers)

            if response.status_code != 200:
                raise ToolExecutionError(
                    message="The page or database could not be found.",
                    developer_message=f"The response was: {response.json()}",
                )

            return dict(response.json())

    if object_id is not None and object_id != "":
        return await get_metadata_by_id(object_id)
    elif object_title is not None and object_title != "":
        return await get_metadata_by_title(object_title)
    else:
        raise ToolExecutionError(
            message="Either object_title or object_id must be provided.",
        )


@tool(requires_auth=Notion())
async def get_workspace_structure(
    context: ToolContext,
) -> Annotated[dict[str, Any], "The workspace structure"]:
    """Get the workspace structure of the user's Notion workspace.
    Ideal for finding where an object is located in the workspace.
    """
    # Retrieve the complete flat list of all pages and databases.
    results = await search_by_title(context, None, limit=-1)

    # Remove database rows from results
    # They're returned from the search results because they're
    # technically child pages of the database, but since they're not displayed in the UI's
    # sidebar workspace structure, we do not include them in this tool's response.
    results["results"] = [
        item
        for item in results.get("results", [])
        if not (
            item.get("object", "") == "page"
            and item.get("parent", {}).get("type", "") == "database_id"
        )
    ]

    async with httpx.AsyncClient() as client:
        headers = get_headers(context)
        orphaned_items = []
        for item in results.get("results", []):
            # This condition will only be met for databases that are 'child_pages' of a page.
            # Notion API wraps these databases in a block object, so we need to unwrap it to
            # link the parent page to the database. Sometimes it takes multiple unwrappings
            # to get to the parent page.
            while (
                item.get("parent", {}).get("type", "") == "block_id"
                and item.get("type", "database") == "database"
            ):
                parent = item.get("parent", {})
                block_id = parent["block_id"]
                url = get_url("retrieve_a_block", block_id=block_id)
                block_response = await client.get(url, headers=headers)
                if block_response.status_code != 200:
                    # unable to attach the database to the parent page
                    orphaned_items.append(item["id"])
                    break
                block_data = block_response.json()
                if "parent" in block_data:
                    item["parent"] = block_data["parent"]

        # Drop orphaned items from results since we were unable to attach them to a parent page.
        results["results"] = [
            item for item in results.get("results", []) if item["id"] not in orphaned_items
        ]

    items = results.get("results", [])
    workspace_tree = build_workspace_structure(items)

    return workspace_tree



from typing import Annotated, Any

from arcade_tdk import ToolContext, tool

from arcade_google_flights.enums import (
    GoogleFlightsMaxStops,
    GoogleFlightsSortBy,
    GoogleFlightsTravelClass,
)
from arcade_google_flights.utils import call_serpapi, parse_flight_results, prepare_params


@tool(requires_secrets=["SERP_API_KEY"])
async def search_one_way_flights(
    context: ToolContext,
    departure_airport_code: Annotated[
        str, "The departure airport code. An uppercase 3-letter code"
    ],
    arrival_airport_code: Annotated[str, "The arrival airport code. An uppercase 3-letter code"],
    outbound_date: Annotated[str, "Flight departure date in YYYY-MM-DD format"],
    currency_code: Annotated[
        str | None, "Currency of the returned prices. Defaults to 'USD'"
    ] = "USD",
    travel_class: Annotated[
        GoogleFlightsTravelClass,
        "Travel class of the flight. Defaults to 'ECONOMY'",
    ] = GoogleFlightsTravelClass.ECONOMY,
    num_adults: Annotated[int | None, "Number of adult passengers. Defaults to 1"] = 1,
    num_children: Annotated[int | None, "Number of child passengers. Defaults to 0"] = 0,
    max_stops: Annotated[
        GoogleFlightsMaxStops,
        "Maximum number of stops (layovers) for the flight. Defaults to any number of stops",
    ] = GoogleFlightsMaxStops.ANY,
    sort_by: Annotated[
        GoogleFlightsSortBy,
        "The sorting order of the results. Defaults to TOP_FLIGHTS.",
    ] = GoogleFlightsSortBy.TOP_FLIGHTS,
) -> Annotated[dict[str, Any], "Flight search results from the Google Flights API"]:
    """Retrieve flight search results for a one-way flight using Google Flights"""
    params = prepare_params(
        "google_flights",
        departure_id=departure_airport_code,
        arrival_id=arrival_airport_code,
        outbound_date=outbound_date,
        currency=currency_code,
        travel_class=travel_class.to_api_value(),
        adults=num_adults,
        children=num_children,
        stops=max_stops.to_api_value(),
        sort_by=sort_by.to_api_value(),
        type=2,  # indicates one-way
        deep_search=True,  # Same search depth as the Google Flights page in the browser
    )

    # Execute the request
    results = call_serpapi(context, params)

    # Parse the results
    flights = parse_flight_results(results)

    return flights



from typing import Annotated

import pytest
from arcade_core.catalog import ToolCatalog
from arcade_core.executor import ToolExecutor
from arcade_core.schema import ToolCallError, ToolCallLog, ToolCallOutput, ToolContext
from arcade_tdk import tool
from arcade_tdk.errors import RetryableToolError, ToolExecutionError


@tool
def simple_tool(inp: Annotated[str, "input"]) -> Annotated[str, "output"]:
    """Simple tool"""
    return inp


@tool.deprecated("Use simple_tool instead")
@tool
def simple_deprecated_tool(inp: Annotated[str, "input"]) -> Annotated[str, "output"]:
    """Simple tool that is deprecated"""
    return inp


@tool
def retryable_error_tool() -> Annotated[str, "output"]:
    """Tool that raises a retryable error"""
    raise RetryableToolError("test", "test", "test", 1000)


@tool
def exec_error_tool() -> Annotated[str, "output"]:
    """Tool that raises an error"""
    raise ToolExecutionError("test", "test")


@tool
def unexpected_error_tool() -> Annotated[str, "output"]:
    """Tool that raises an unexpected error"""
    raise RuntimeError("test")


@tool
def bad_output_error_tool() -> Annotated[str, "output"]:
    """tool that returns a bad output type"""
    return {"output": "test"}


# ---- Test Driver ----

catalog = ToolCatalog()
catalog.add_tool(simple_tool, "simple_toolkit")
catalog.add_tool(simple_deprecated_tool, "simple_toolkit")
catalog.add_tool(retryable_error_tool, "simple_toolkit")
catalog.add_tool(exec_error_tool, "simple_toolkit")
catalog.add_tool(unexpected_error_tool, "simple_toolkit")
catalog.add_tool(bad_output_error_tool, "simple_toolkit")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "tool_func, inputs, expected_output",
    [
        (simple_tool, {"inp": "test"}, ToolCallOutput(value="test")),
        (
            simple_deprecated_tool,
            {"inp": "test"},
            ToolCallOutput(
                value="test",
                logs=[
                    ToolCallLog(
                        message="Use simple_tool instead",
                        level="warning",
                        subtype="deprecation",
                    )
                ],
            ),
        ),
        (
            retryable_error_tool,
            {},
            ToolCallOutput(
                error=ToolCallError(
                    message="test",
                    developer_message="test",
                    additional_prompt_content="test",
                    retry_after_ms=1000,
                    can_retry=True,
                )
            ),
        ),
        (
            exec_error_tool,
            {},
            ToolCallOutput(
                error=ToolCallError(
                    message="test",
                    developer_message="test",
                )
            ),
        ),
        (
            unexpected_error_tool,
            {},
            ToolCallOutput(
                error=ToolCallError(
                    message="Error in execution of UnexpectedErrorTool",
                    developer_message="Error in unexpected_error_tool: test",
                )
            ),
        ),
        (
            simple_tool,
            {"inp": {"test": "test"}},  # takes in a string not a dict
            ToolCallOutput(
                error=ToolCallError(
                    message="Error in tool input deserialization",
                    developer_message=None,  # can't gaurantee this will be the same
                )
            ),
        ),
        (
            bad_output_error_tool,
            {},
            ToolCallOutput(
                error=ToolCallError(
                    message="Failed to serialize tool output",
                    developer_message=None,  # can't gaurantee this will be the same
                )
            ),
        ),
    ],
    ids=[
        "simple_tool",
        "simple_deprecated_tool",
        "retryable_error_tool",
        "exec_error_tool",
        "unexpected_error_tool",
        "invalid_input_type",
        "bad_output_type",
    ],
)
async def test_tool_executor(tool_func, inputs, expected_output):
    tool_definition = catalog.find_tool_by_func(tool_func)

    dummy_context = ToolContext()
    full_tool = catalog.get_tool(tool_definition.get_fully_qualified_name())
    output = await ToolExecutor.run(
        func=tool_func,
        definition=tool_definition,
        input_model=full_tool.input_model,
        output_model=full_tool.output_model,
        context=dummy_context,
        **inputs,
    )

    check_output(output, expected_output)


def check_output(output: ToolCallOutput, expected_output: ToolCallOutput):
    # execution error in tool
    if output.error:
        assert output.error.message == expected_output.error.message
        if expected_output.error.developer_message:
            assert output.error.developer_message == expected_output.error.developer_message
        if expected_output.error.traceback_info:
            assert output.error.traceback_info == expected_output.error.traceback_info
        assert output.error.can_retry == expected_output.error.can_retry
        assert (
            output.error.additional_prompt_content
            == expected_output.error.additional_prompt_content
        )
        assert output.error.retry_after_ms == expected_output.error.retry_after_ms

    # normal tool execution
    else:
        assert output.value == expected_output.value

        # check logs
        output_logs = output.logs or []
        expected_logs = expected_output.logs or []
        assert len(output_logs) == len(expected_logs)
        for output_log, expected_log in zip(output_logs, expected_logs):
            assert output_log.message == expected_log.message
            assert output_log.level == expected_log.level
            assert output_log.subtype == expected_log.subtype



SPOTIFY_BASE_URL = "https://api.spotify.com/v1"

ENDPOINTS = {
    "player_get_available_devices": "/me/player/devices",
    "player_get_playback_state": "/me/player",
    "player_get_currently_playing": "/me/player/currently-playing",
    "player_modify_playback": "/me/player/play",
    "player_pause_playback": "/me/player/pause",
    "player_skip_to_next": "/me/player/next",
    "player_skip_to_previous": "/me/player/previous",
    "player_seek_to_position": "/me/player/seek",
    "tracks_get_track": "/tracks/{track_id}",
    "search": "/search",
}

RESPONSE_MSGS = {
    "artist_not_found": "Artist '{artist_name}' not found",
    "track_not_found": "Track '{track_name}' not found",
    "no_track_to_adjust_position": "No track to adjust position",
    "playback_position_adjusted": "Playback position adjusted",
    "no_track_to_go_back_to": "No track to go back to",
    "playback_skipped_to_previous_track": "Playback skipped to previous track",
    "no_track_to_skip": "No track to skip",
    "playback_skipped_to_next_track": "Playback skipped to next track",
    "playback_paused": "Playback paused",
    "playback_resumed": "Playback resumed",
    "no_track_to_resume": "No track to resume",
    "no_track_to_pause": "No track to pause",
    "no_track_to_play": "No track to play",
    "no_available_devices": "No available devices",
    "track_is_already_paused": "Track is already paused",
    "track_is_already_playing": "Track is already playing",
    "playback_started": "Playback started",
    "no_active_device": "Cannot start playback because no active device is available",
}



import asyncio
import base64
import json
import mimetypes
from collections.abc import Callable
from contextlib import suppress
from datetime import date, datetime
from typing import Any, cast

from arcade_tdk import ToolContext
from arcade_tdk.errors import ToolExecutionError

from arcade_jira.constants import STOP_WORDS
from arcade_jira.exceptions import JiraToolExecutionError, MultipleItemsFoundError, NotFoundError


def remove_none_values(data: dict) -> dict:
    """Remove all keys with None values from the dictionary."""
    return {k: v for k, v in data.items() if v is not None}


def safe_delete_dict_keys(data: dict, keys: list[str]) -> dict:
    for key in keys:
        with suppress(KeyError):
            del data[key]
    return data


def convert_date_string_to_date(date_string: str) -> date:
    return datetime.strptime(date_string, "%Y-%m-%d").date()


def is_valid_date_string(date_string: str) -> bool:
    try:
        convert_date_string_to_date(date_string)
    except ValueError:
        return False

    return True


def quote(v: str) -> str:
    quoted = v.replace('"', r"\"")
    return f'"{quoted}"'


def build_search_issues_jql(
    keywords: str | None = None,
    due_from: date | None = None,
    due_until: date | None = None,
    status: str | None = None,
    priority: str | None = None,
    assignee: str | None = None,
    project: str | None = None,
    issue_type: str | None = None,
    labels: list[str] | None = None,
    parent_issue: str | None = None,
) -> str:
    clauses: list[str] = []

    if keywords:
        kw_clauses = [
            f"text ~ {quote(k.casefold())}"
            for k in keywords.split()
            if k.casefold() not in STOP_WORDS
        ]
        clauses.append("(" + " AND ".join(kw_clauses) + ")")

    if due_from:
        clauses.append(f'dueDate >= "{due_from.isoformat()}"')
    if due_until:
        clauses.append(f'dueDate <= "{due_until.isoformat()}"')

    if labels:
        label_list = ",".join(quote(label) for label in labels)
        clauses.append(f"labels IN ({label_list})")

    standard_cases = [
        ("status", status),
        ("priority", priority),
        ("assignee", assignee),
        ("project", project),
        ("issuetype", issue_type),
        ("parent", parent_issue),
    ]

    for field, value in standard_cases:
        if value:
            clauses.append(f"{field} = {quote(value)}")

    return " AND ".join(clauses) if clauses else ""


def clean_issue_dict(issue: dict, cloud_name: str | None = None) -> dict:
    fields = cast(dict, issue["fields"])
    rendered_fields = issue.get("renderedFields", {})

    fields["id"] = issue["id"]
    fields["key"] = issue["key"]
    fields["title"] = fields["summary"]

    if fields.get("parent"):
        fields["parent"] = get_summarized_issue_dict(fields["parent"])

    if fields["assignee"]:
        fields["assignee"] = clean_user_dict(fields["assignee"], cloud_name)

    if fields["creator"]:
        fields["creator"] = clean_user_dict(fields["creator"], cloud_name)

    if fields["reporter"]:
        fields["reporter"] = clean_user_dict(fields["reporter"], cloud_name)

    if fields.get("description"):
        fields["description"] = rendered_fields.get("description")

    if fields.get("environment"):
        fields["environment"] = rendered_fields.get("environment")

    if fields.get("worklog"):
        fields["worklog"] = {
            "items": rendered_fields.get("worklog", {}).get("worklogs", []),
            "total": len(rendered_fields.get("worklog", {}).get("worklogs", [])),
        }

    if fields.get("attachment"):
        fields["attachments"] = [
            clean_attachment_dict(attachment, cloud_name)
            for attachment in fields.get("attachment", [])
        ]

    add_identified_fields_to_issue(fields, ["status", "issuetype", "priority", "project"])

    safe_delete_dict_keys(
        fields,
        [
            "subtasks",
            "summary",
            "assignee",
            "creator",
            "issuetype",
            "lastViewed",
            "updated",
            "statusCategory",
            "statuscategorychangedate",
            "votes",
            "watches",
            "attachment",
            "comment",
            "self",
        ],
    )

    fields["url"] = build_issue_url(cloud_name, fields["project"]["key"], fields["key"])

    return fields


def add_identified_fields_to_issue(
    fields_dict: dict[str, Any],
    field_names: list[str],
) -> dict[str, Any]:
    for field_name in field_names:
        if fields_dict.get(field_name):
            data = {
                "name": fields_dict[field_name]["name"],
                "id": fields_dict[field_name]["id"],
            }
            if "key" in fields_dict[field_name]:
                data["key"] = fields_dict[field_name]["key"]
            fields_dict[field_name] = data

    return fields_dict


def clean_comment_dict(comment: dict, include_adf_content: bool = False) -> dict:
    data = {
        "id": comment["id"],
        "author": {
            "name": comment["author"]["displayName"],
            "email": comment["author"]["emailAddress"],
        },
        "body": comment["renderedBody"],
        "created_at": comment["created"],
    }

    if include_adf_content:
        data["adf_body"] = comment["body"]

    return data


def clean_project_dict(project: dict, cloud_name: str | None = None) -> dict:
    data = {
        "id": project["id"],
        "key": project["key"],
        "name": project["name"],
    }

    data["url"] = build_project_url(cloud_name, project["key"])

    if "description" in project:
        data["description"] = project["description"]

    if "email" in project:
        data["email"] = project["email"]

    if "projectCategory" in project:
        data["category"] = project["projectCategory"]

    if "style" in project:
        data["style"] = project["style"]

    return data


def clean_issue_type_dict(issue_type: dict) -> dict:
    data = {
        "id": issue_type["id"],
        "name": issue_type["name"],
        "description": issue_type["description"],
    }

    if "scope" in issue_type:
        data["scope"] = issue_type["scope"]

    return data


def clean_user_dict(user: dict, cloud_name: str | None = None) -> dict:
    data = {
        "id": user["accountId"],
        "name": user["displayName"],
        "active": user["active"],
    }

    data["url"] = build_user_url(cloud_name, user["accountId"])

    if user.get("emailAddress"):
        data["email"] = user["emailAddress"]

    if user.get("accountType"):
        data["account_type"] = user["accountType"]

    if user.get("timeZone"):
        data["timezone"] = user["timeZone"]

    if user.get("active"):
        data["active"] = user["active"]

    return data


def clean_attachment_dict(attachment: dict, cloud_name: str | None = None) -> dict:
    return {
        "id": attachment["id"],
        "filename": attachment["filename"],
        "mime_type": attachment["mimeType"],
        "size": {"bytes": attachment["size"]},
        "author": clean_user_dict(attachment["author"], cloud_name),
    }


def clean_priority_scheme_dict(scheme: dict, cloud_name: str | None = None) -> dict:
    data = {
        "id": scheme["id"],
        "name": scheme["name"],
        "description": scheme["description"],
        "is_default": scheme["isDefault"],
    }

    if isinstance(scheme.get("priorities"), dict):
        all_priorities = scheme["priorities"].get("isLast", True)

        data["priorities"] = [
            clean_priority_dict(priority) for priority in scheme["priorities"]["values"]
        ]

        if not all_priorities:
            # Avoid circular import
            from arcade_jira.tools.priorities import (
                list_priorities_associated_with_a_priority_scheme,
            )

            data["priorities"]["message"] = (
                "Not all priorities are listed. Paginate the "
                f"`Jira.{list_priorities_associated_with_a_priority_scheme.__tool_name__}` tool "
                "to get the full list of priorities in this priority scheme."
            )

    if isinstance(scheme.get("projects"), dict):
        all_projects = scheme["projects"].get("isLast", True)
        data["projects"] = [
            clean_project_dict(project, cloud_name) for project in scheme["projects"]["values"]
        ]
        if not all_projects:
            # Avoid circular import
            from arcade_jira.tools.priorities import list_projects_associated_with_a_priority_scheme

            data["projects"]["message"] = (
                "Not all projects are listed. Paginate the "
                f"`Jira.{list_projects_associated_with_a_priority_scheme.__tool_name__}` tool "
                "to get the full list of projects in this priority scheme."
            )

    return data


def clean_priority_dict(priority: dict) -> dict:
    data = {
        "id": priority["id"],
        "name": priority["name"],
        "description": priority["description"],
    }

    if "statusColor" in priority:
        data["statusColor"] = priority["statusColor"]

    return data


def clean_labels(labels: list[str] | None) -> list[str] | None:
    if not labels:
        return None
    return [label.strip().replace(" ", "_") for label in labels]


def get_summarized_issue_dict(issue: dict) -> dict:
    fields = issue["fields"]
    return {
        "id": issue["id"],
        "key": issue["key"],
        "title": fields.get("summary"),
        "status": fields.get("status", {}).get("name"),
        "type": fields.get("issuetype", {}).get("name"),
        "priority": fields.get("priority", {}).get("name"),
    }


def add_pagination_to_response(
    response: dict[str, Any],
    items: list[dict[str, Any]],
    limit: int,
    offset: int,
    max_results: int | None = None,
) -> dict[str, Any]:
    next_offset = offset + limit
    if max_results:
        next_offset = min(next_offset, max_results - limit)

    response["pagination"] = {
        "limit": limit,
        "total_results": len(items),
    }

    if response.get("isLast") is True:
        response["pagination"]["is_last_page"] = True
    elif response.get("isLast") is False or (len(items) >= limit and next_offset > offset):
        response["pagination"]["next_offset"] = next_offset
    else:
        response["pagination"]["is_last_page"] = True

    with suppress(KeyError):
        del response["isLast"]

    return response


def simplify_user_dict(user: dict) -> dict:
    return {
        "id": user["id"],
        "name": user["name"],
        "email": user["email"],
    }


async def find_multiple_unique_users(
    context: ToolContext,
    user_identifiers: list[str],
    exact_match: bool = False,
) -> list[dict[str, Any]]:
    """
    Find users matching either their display name, email address, or account ID.

    By default, the search will match prefixes. A user_identifier of "john" will match
    "John Doe", "Johnson", "john.doe@example.com", etc.

    If `enforce_exact_match` is set to True, the search will only return users that have either
    a display name, email address, or account ID that match the exact user_identifier.
    """
    from arcade_jira.tools.users import (  # Avoid circular import
        get_user_by_id,
        get_users_without_id,
    )

    users: list[dict[str, Any]] = []

    responses = await asyncio.gather(*[
        get_users_without_id(
            context=context,
            name_or_email=user_identifier,
            enforce_exact_match=exact_match,
        )
        for user_identifier in user_identifiers
    ])

    search_by_id: list[str] = []

    for response in responses:
        user_identifier = response["query"]["name_or_email"]

        if response["pagination"]["total_results"] > 1:
            simplified_users = [simplify_user_dict(user) for user in response["users"]]
            raise MultipleItemsFoundError(
                message=f"Multiple users found with name or email '{user_identifier}'. "
                f"Please provide a unique ID: {json.dumps(simplified_users)}"
            )

        elif response["pagination"]["total_results"] == 0:
            search_by_id.append(user_identifier)

        else:
            users.append(response["users"][0])

    if search_by_id:
        responses = await asyncio.gather(*[
            get_user_by_id(context, user_id=user_id) for user_id in search_by_id
        ])
        for response in responses:
            if response["user"]:
                users.append(response["user"])
            else:
                raise NotFoundError(
                    message=f"No user found with '{response['query']['user_id']}'.",
                )

    return users


async def find_unique_project(
    context: ToolContext,
    project_identifier: str,
) -> dict[str, Any]:
    """Find a unique project by its ID, key, or name

    Args:
        project_identifier: The ID, key, or name of the project to find.

    Returns:
        The project found.
    """
    # Avoid circular import
    from arcade_jira.tools.projects import get_project_by_id, search_projects

    # Try to find project by ID or key
    response = await get_project_by_id(context, project=project_identifier)
    if response.get("project"):
        return cast(dict, response["project"])

    # If not found, search by name
    response = await search_projects(context, keywords=project_identifier)
    projects = response["projects"]
    if len(projects) == 1:
        return cast(dict, projects[0])
    elif len(projects) > 1:
        simplified_projects = [
            {
                "id": project["id"],
                "name": project["name"],
            }
            for project in projects
        ]
        raise MultipleItemsFoundError(
            message=f"Multiple projects found with name/key/ID '{project_identifier}'. "
            f"Please provide a unique ID: {json.dumps(simplified_projects)}"
        )

    raise NotFoundError(message=f"Project not found with name/key/ID '{project_identifier}'")


async def find_unique_priority(
    context: ToolContext,
    priority_identifier: str,
    project_id: str,
) -> dict[str, Any]:
    """Find a unique priority by ID or name that is associated with a project

    Args:
        priority_identifier: The ID or name of the priority to find.
        project_id: The ID of the project to find the priority for.

    Returns:
        The priority found.
    """
    # Avoid circular import
    from arcade_jira.tools.priorities import (
        get_priority_by_id,
        list_priorities_available_to_a_project,
    )

    # Try to get the priority by ID first
    response = await get_priority_by_id(context, priority_identifier)
    if response.get("priority"):
        return cast(dict, response["priority"])

    # If not found, search by name
    response = await list_priorities_available_to_a_project(context, project_id)

    if response.get("error"):
        raise JiraToolExecutionError(response["error"])

    priorities = response["priorities_available"]
    matches: list[dict[str, Any]] = []

    for priority in priorities:
        if priority["name"].casefold() == priority_identifier.casefold():
            matches.append(priority)

    if len(matches) == 1:
        return cast(dict, matches[0])
    elif len(matches) > 1:
        simplified_matches = [
            {
                "id": match["id"],
                "name": match["name"],
            }
            for match in matches
        ]
        raise MultipleItemsFoundError(
            message=f"Multiple priorities found with name '{priority_identifier}'. "
            f"Please provide a unique ID: {json.dumps(simplified_matches)}"
        )

    raise NotFoundError(message=f"Priority not found with ID or name '{priority_identifier}'")


async def find_unique_issue_type(
    context: ToolContext,
    issue_type_identifier: str,
    project_id: str,
) -> dict[str, Any]:
    """Find a unique issue type by its ID or name that is associated with a project

    Args:
        issue_type_identifier: The ID or name of the issue type to find.
        project_id: The ID of the project to find the issue type for.

    Returns:
        The issue type found.
    """
    # Avoid circular import
    from arcade_jira.tools.issues import get_issue_type_by_id, list_issue_types_by_project

    # Try to get the issue type by ID first
    response = await get_issue_type_by_id(context, issue_type_identifier)
    if response.get("issue_type"):
        return cast(dict, response["issue_type"])

    # If not found, search by name
    response = await list_issue_types_by_project(context, project_id)

    if response.get("error"):
        raise JiraToolExecutionError(response["error"])

    issue_types = response["issue_types"]
    matches: list[dict[str, Any]] = []

    for issue_type in issue_types:
        if issue_type["name"].casefold() == issue_type_identifier.casefold():
            matches.append(issue_type)

    if len(matches) == 1:
        return cast(dict, matches[0])
    elif len(matches) > 1:
        simplified_matches = [
            {
                "id": match["id"],
                "name": match["name"],
            }
            for match in matches
        ]
        raise MultipleItemsFoundError(
            message=f"Multiple issue types found with name '{issue_type_identifier}'. "
            f"Please provide a unique ID: {json.dumps(simplified_matches)}"
        )

    available_issue_types = json.dumps([
        {
            "id": issue_type["id"],
            "name": issue_type["name"],
        }
        for issue_type in issue_types
    ])

    raise NotFoundError(
        message=f"Issue type not found with ID or name '{issue_type_identifier}'. "
        f"These are the issue types available for the project: {available_issue_types}"
    )


async def find_unique_user(
    context: ToolContext,
    user_identifier: str,
) -> dict[str, Any]:
    """Find a unique user by their ID, key, email address, or display name."""
    # Avoid circular import
    from arcade_jira.tools.users import get_user_by_id, get_users_without_id

    # Try to get the user by ID
    response = await get_user_by_id(context, user_identifier)
    if response.get("user"):
        return cast(dict, response["user"])

    # Search for the user name or email, if not found by ID
    response = await get_users_without_id(
        context, name_or_email=user_identifier, enforce_exact_match=True
    )
    users = response["users"]

    if len(users) == 1:
        return cast(dict, users[0])
    elif len(users) > 1:
        simplified_users = [
            {
                "id": user["id"],
                "name": user["name"],
                "email": user["email"],
            }
            for user in users
        ]
        raise MultipleItemsFoundError(
            message=f"Multiple users found with name or email '{user_identifier}'. "
            f"Please provide a unique ID: {json.dumps(simplified_users)}"
        )

    raise NotFoundError(message=f"User not found with ID, name or email '{user_identifier}'")


async def get_single_project(context: ToolContext) -> dict[str, Any]:
    from arcade_jira.tools.projects import list_projects

    projects = await paginate_all_items(
        context=context,
        tool=list_projects,
        response_items_key="projects",
    )

    if len(projects) == 0:
        raise NotFoundError(message="No projects found in this account.")

    if len(projects) == 1:
        return cast(dict[str, Any], projects[0])

    available_projects_str = json.dumps([
        {
            "id": project["id"],
            "name": project["name"],
        }
        for project in projects
    ])

    raise MultipleItemsFoundError(message=f"Multiple projects found: {available_projects_str}")


def build_file_data(
    filename: str,
    file_content_str: str | None,
    file_content_base64: str | None,
    file_type: str | None = None,
    file_encoding: str = "utf-8",
) -> dict[str, tuple]:
    if file_content_str is not None:
        try:
            file_content = file_content_str.encode(file_encoding)
        except LookupError as exc:
            raise ToolExecutionError(message=f"Unknown encoding: {file_encoding}") from exc
        except Exception as exc:
            raise ToolExecutionError(
                message=f"Failed to encode file content string with {file_encoding} "
                f"encoding: {exc!s}"
            ) from exc
    elif file_content_base64 is not None:
        try:
            file_content = base64.b64decode(file_content_base64)
        except Exception as exc:
            raise ToolExecutionError(
                message=f"Failed to decode base64 file content: {exc!s}"
            ) from exc

    if not file_type:
        # guess_type returns None if the file type is not recognized
        file_type = mimetypes.guess_type(filename)[0]

    if file_type:
        return {"file": (filename, file_content, file_type)}

    return {"file": (filename, file_content)}


def build_adf_doc(text: str) -> dict:
    return {
        "type": "doc",
        "version": 1,
        "content": [
            {
                "type": "paragraph",
                "content": [{"type": "text", "text": text}],
            }
            for text in text.split("\n")
        ],
    }


async def paginate_all_items(
    context: ToolContext,
    tool: Callable,
    response_items_key: str,
    limit: int | None = None,
    offset: int | None = None,
    **kwargs: Any,
) -> list[Any]:
    """Paginate all items from a tool."""
    keep_paginating = True
    items: list[Any] = []

    if limit is not None:
        kwargs["limit"] = limit

    if offset is not None:
        kwargs["offset"] = offset

    while keep_paginating:
        response = await tool(context, **kwargs)

        if response.get("error"):
            raise JiraToolExecutionError(response["error"])

        next_offset = response["pagination"].get("next_offset")
        kwargs["offset"] = next_offset
        keep_paginating = isinstance(next_offset, int)
        items.extend(response[response_items_key])

    return items


async def paginate_all_priority_schemes(context: ToolContext) -> list[dict]:
    """Get all priority schemes."""
    # Avoid circular import
    from arcade_jira.tools.priorities import list_priority_schemes

    return await paginate_all_items(context, list_priority_schemes, "priority_schemes")


async def paginate_all_priorities_by_priority_scheme(
    context: ToolContext,
    scheme_id: str,
) -> list[dict]:
    """Get all priorities associated with a priority scheme."""
    # Avoid circular import
    from arcade_jira.tools.priorities import list_priorities_associated_with_a_priority_scheme

    return await paginate_all_items(
        context,
        list_priorities_associated_with_a_priority_scheme,
        "priorities",
        scheme_id=scheme_id,
    )


async def paginate_all_issue_types(context: ToolContext, project_identifier: str) -> list[dict]:
    """Get all issue types associated with a project."""
    # Avoid circular import
    from arcade_jira.tools.issues import list_issue_types_by_project

    return await paginate_all_items(
        context,
        list_issue_types_by_project,
        "issue_types",
        project=project_identifier,
    )


async def validate_issue_args(
    context: ToolContext,
    due_date: str | None,
    project: str | None,
    issue_type: str | None,
    priority: str | None,
    parent_issue: str | None,
) -> tuple[dict | None, dict | None, str | dict | None, str | dict | None, dict | None]:
    if due_date and not is_valid_date_string(due_date):
        return (
            {"error": f"Invalid `due_date` format: '{due_date}'. Please use YYYY-MM-DD."},
            None,
            None,
            None,
            None,
        )

    if not project and not parent_issue:
        return (
            {"error": "Must provide either `project` or `parent_issue` argument."},
            None,
            None,
            None,
            None,
        )

    error: dict[str, Any] | None = None
    project_data = await get_project_by_project_identifier_or_by_parent_issue(
        context, project, parent_issue
    )
    issue_type_data: str | dict[str, Any] | None = None
    priority_data: str | dict[str, Any] | None = None
    parent_issue_data: dict[str, Any] | None = None

    if project_data.get("error"):
        error = project_data
        return error, None, issue_type_data, priority_data, parent_issue_data

    error, issue_type_data = await resolve_issue_type(context, issue_type, project_data)
    if error:
        return error, project_data, issue_type_data, priority_data, parent_issue_data

    error, priority_data = await resolve_issue_priority(context, priority, project_data)
    if error:
        return error, project_data, issue_type_data, priority_data, parent_issue_data

    error, parent_issue_data = await resolve_parent_issue(context, parent_issue)
    if error:
        return error, project_data, issue_type_data, priority_data, parent_issue_data

    return None, project_data, issue_type_data, priority_data, parent_issue_data


async def resolve_issue_type(
    context: ToolContext,
    issue_type: str | None,
    project_data: dict,
) -> tuple[dict[str, Any] | None, str | dict[str, Any] | None]:
    if issue_type == "":
        return None, ""
    elif issue_type:
        try:
            response = await find_unique_issue_type(context, issue_type, project_data["id"])
        except JiraToolExecutionError as exc:
            return {"error": exc.message}, None
        else:
            return None, response

    return None, None


async def resolve_issue_priority(
    context: ToolContext,
    priority: str | None,
    project_data: dict,
) -> tuple[dict[str, Any] | None, str | dict[str, Any] | None]:
    if priority == "":
        return None, ""
    elif priority:
        try:
            priority_data = await find_unique_priority(context, priority, project_data["id"])
        except JiraToolExecutionError as exc:
            return {"error": exc.message}, None
        else:
            return None, priority_data

    return None, None


async def resolve_parent_issue(
    context: ToolContext,
    parent_issue: str | None,
) -> tuple[dict[str, Any] | None, dict[str, Any] | None]:
    if parent_issue == "":
        return {"error": "Parent issue cannot be empty"}, None
    elif parent_issue:
        from arcade_jira.tools.issues import get_issue_by_id  # Avoid circular import

        try:
            parent_issue_data = await get_issue_by_id(context, parent_issue)
        except JiraToolExecutionError as exc:
            return {"error": exc.message}, None
        else:
            return None, parent_issue_data["issue"]

    return None, None


async def get_project_by_project_identifier_or_by_parent_issue(
    context: ToolContext,
    project: str | None,
    parent_issue_id: str | None,
) -> dict[str, Any]:
    from arcade_jira.tools.issues import get_issue_by_id  # Avoid circular import

    if not project and not parent_issue_id:
        return {"error": "Must provide either `project` or `parent_issue_id` argument."}

    if not project:
        parent_issue_data = await get_issue_by_id(context, parent_issue_id)
        if parent_issue_data.get("error"):
            return {"error": f"Parent issue not found with ID {parent_issue_id}."}
        project = cast(str, parent_issue_data["project"]["id"])

    try:
        project_data = await find_unique_project(context, project)
    except JiraToolExecutionError as exc:
        return {"error": exc.message}

    return project_data


async def resolve_issue_users(
    context: ToolContext,
    assignee: str | None,
    reporter: str | None,
) -> tuple[dict | None, str | dict | None, str | dict | None]:
    assignee_data: str | dict | None = None
    reporter_data: str | dict | None = None

    if (not assignee and assignee != "") and (not reporter and reporter != ""):
        return None, None, None

    if assignee == "":
        assignee_data = ""
    elif assignee:
        try:
            assignee_data = await find_unique_user(context, assignee)
        except JiraToolExecutionError as exc:
            return {"error": exc.message}, assignee_data, reporter_data

    if reporter == "":
        reporter_data = ""
    elif reporter:
        try:
            reporter_data = await find_unique_user(context, reporter)
        except JiraToolExecutionError as exc:
            return {"error": exc.message}, assignee_data, reporter_data

    return None, assignee_data, reporter_data


async def find_priorities_by_project(
    context: ToolContext,
    project: dict[str, Any],
) -> dict[str, Any]:
    # Avoid circular import
    from arcade_jira.tools.priorities import list_projects_associated_with_a_priority_scheme

    scheme_ids: set[str] = set()
    priority_ids: set[str] = set()
    priorities: list[dict[str, Any]] = []

    priority_schemes = await paginate_all_priority_schemes(context)

    if not priority_schemes:
        raise NotFoundError("No priority schemes found")  # noqa: TRY003

    projects_by_scheme = await asyncio.gather(*[
        list_projects_associated_with_a_priority_scheme(
            context=context,
            scheme_id=scheme["id"],
            project=project["id"],
        )
        for scheme in priority_schemes
    ])

    for scheme_index, scheme_projects in enumerate(projects_by_scheme):
        if scheme_projects.get("error"):
            return cast(dict, scheme_projects)

        for scheme_project in scheme_projects["projects"]:
            if scheme_project["id"] == project["id"]:
                scheme = priority_schemes[scheme_index]
                scheme_ids.add(scheme["id"])
                break

    if not scheme_ids:
        return {"error": f"No priority schemes found for the project {project['id']}"}

    priorities_by_scheme = await asyncio.gather(*[
        paginate_all_priorities_by_priority_scheme(context, scheme_id) for scheme_id in scheme_ids
    ])

    for priorities_available in priorities_by_scheme:
        for priority in priorities_available:
            if priority["id"] in priority_ids:
                continue
            priority_ids.add(priority["id"])
            priorities.append(priority)

    return {
        "project": {
            "id": project["id"],
            "key": project["key"],
            "name": project["name"],
        },
        "priorities_available": priorities,
    }


def build_issue_update_request_body(
    title: str | None,
    description: str | None,
    environment: str | None,
    due_date: str | None,
    parent_issue: dict | None,
    issue_type: str | dict | None,
    priority: str | dict | None,
    assignee: str | dict | None,
    reporter: str | dict | None,
    labels: list[str] | None,
) -> dict[str, Any]:
    body: dict[str, dict[str, Any]] = {"fields": {}, "update": {}}

    build_issue_update_text_fields(body, title, description, environment)
    build_issue_update_classifier_fields(body, issue_type, priority)
    build_issue_update_user_fields(body, assignee, reporter)
    build_issue_update_hierarchy_fields(body, parent_issue)
    build_issue_update_date_fields(body, due_date)

    if labels == []:
        body["update"]["labels"] = [{"set": None}]
    elif labels:
        body["fields"]["labels"] = labels

    return body


def build_issue_update_text_fields(
    body: dict,
    title: str | None,
    description: str | None,
    environment: str | None,
) -> dict[str, dict[str, Any]]:
    if title == "":
        raise ValueError("Title cannot be empty")  # noqa: TRY003
    elif title:
        body["fields"]["summary"] = title

    if description == "":
        body["update"]["description"] = [{"set": None}]
    elif description:
        body["fields"]["description"] = build_adf_doc(description)

    if environment == "":
        body["update"]["environment"] = [{"set": None}]
    elif environment:
        body["fields"]["environment"] = build_adf_doc(environment)

    return body


def build_issue_update_user_fields(
    body: dict,
    assignee: str | dict | None,
    reporter: str | dict | None,
) -> dict[str, dict[str, Any]]:
    if assignee == "":
        body["update"]["assignee"] = [{"set": None}]
    elif isinstance(assignee, dict):
        body["fields"]["assignee"] = {"id": assignee["id"]}
    elif assignee is not None:
        raise ValueError(f"Invalid assignee: '{assignee}'")  # noqa: TRY003

    if reporter == "":
        body["update"]["reporter"] = [{"set": None}]
    elif isinstance(reporter, dict):
        body["fields"]["reporter"] = {"id": reporter["id"]}
    elif reporter is not None:
        raise ValueError(f"Invalid reporter: '{reporter}'")  # noqa: TRY003

    return body


def build_issue_update_classifier_fields(
    body: dict,
    issue_type: str | dict | None,
    priority: str | dict | None,
) -> dict[str, dict[str, Any]]:
    if issue_type == "":
        raise ValueError("Issue type cannot be empty")  # noqa: TRY003
    elif isinstance(issue_type, dict):
        body["fields"]["issuetype"] = {"id": issue_type["id"]}
    elif issue_type is not None:
        raise ValueError(f"Invalid issue type: '{issue_type}'")  # noqa: TRY003

    if priority == "":
        raise ValueError("Priority cannot be empty")  # noqa: TRY003
    elif isinstance(priority, dict):
        body["fields"]["priority"] = {"id": priority["id"]}
    elif priority is not None:
        raise ValueError(f"Invalid priority: '{priority}'")  # noqa: TRY003

    return body


def build_issue_update_hierarchy_fields(
    body: dict,
    parent_issue: dict | None,
) -> dict[str, dict[str, Any]]:
    if parent_issue:
        body["fields"]["parent"] = {"id": parent_issue["id"]}

    return body


def build_issue_update_date_fields(
    body: dict,
    due_date: str | None,
) -> dict[str, dict[str, Any]]:
    if due_date == "":
        body["update"]["duedate"] = [{"set": None}]
    elif due_date:
        body["fields"]["duedate"] = due_date

    return body


def extract_id(field: Any) -> dict[str, str] | None:
    return {"id": field["id"]} if isinstance(field, dict) else None


def build_issue_url(cloud_name: str | None, issue_id: str, issue_key: str) -> str | None:
    if not cloud_name:
        return None

    return f"https://{cloud_name}.atlassian.net/jira/software/projects/{issue_id}/list?selectedIssue={issue_key}"


def build_project_url(cloud_name: str | None, project_key: str) -> str | None:
    if not cloud_name:
        return None

    return f"https://{cloud_name}.atlassian.net/jira/software/projects/{project_key}/summary"


def build_user_url(cloud_name: str | None, user_id: str) -> str | None:
    if not cloud_name:
        return None

    return f"https://{cloud_name}.atlassian.net/jira/people/{user_id}"



from enum import Enum


class CellErrorType(str, Enum):
    """The type of error in a cell

    Implementation of https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/other#ErrorType
    """

    ERROR_TYPE_UNSPECIFIED = "ERROR_TYPE_UNSPECIFIED"  # The default error type, do not use this.
    ERROR = "ERROR"  # Corresponds to the #ERROR! error.
    NULL_VALUE = "NULL_VALUE"  # Corresponds to the #NULL! error.
    DIVIDE_BY_ZERO = "DIVIDE_BY_ZERO"  # Corresponds to the #DIV/0 error.
    VALUE = "VALUE"  # Corresponds to the #VALUE! error.
    REF = "REF"  # Corresponds to the #REF! error.
    NAME = "NAME"  # Corresponds to the #NAME? error.
    NUM = "NUM"  # Corresponds to the #NUM! error.
    N_A = "N_A"  # Corresponds to the #N/A error.
    LOADING = "LOADING"  # Corresponds to the Loading... state.


class NumberFormatType(str, Enum):
    NUMBER = "NUMBER"
    PERCENT = "PERCENT"
    CURRENCY = "CURRENCY"



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_notion_toolkit
from arcade_notion_toolkit.enums import ObjectType
from arcade_notion_toolkit.tools import (
    get_object_metadata,
    get_workspace_structure,
    search_by_title,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
catalog.add_module(arcade_notion_toolkit)

# Additional messages
TOOL_CALLING_CONVERSATION = [
    {"role": "user", "content": "hi"},
    {"role": "assistant", "content": "Hello! How can I assist you today?"},
    {"role": "user", "content": "explain LLM tool calling"},
    {
        "role": "assistant",
        "content": "Tool calling in the context of Large Language Models (LLMs) refers to the process where the model uses external tools, plugins, orfunctions to gather, process, or enhance information. This allows the model to access more current or specific data and perform actions that are beyond itstrained capabilities.\n\nHere's a breakdown of how this works:\n\n1. **Identification of Needs**: The LLM can identify when a task or query requires dataor actions that are not within its built-in knowledge. This could be due to the data being too recent, too specialized, or involving actions that requireinterfacing with external systems.\n\n2. **Tool Selection**: The LLM selects appropriate tools or functions that are registered and permitted for use. Thisselection is based on the task's requirements, such as retrieving data, performing calculations, or managing content.\n\n3. **Parameter Configuration**:The LLM prepares the necessary input parameters for the tools. These parameters tell the tools what specific actions to take or what data toretrieve.\n\n4. **Execution**: The tools are executed. Some tools run asynchronously, allowing the LLM to perform other tasks while waiting for theresults.\n\n5. **Result Integration**: The LLM integrates the results from the tools back into the conversation or task. It uses this information to answerquestions, perform further analysis, or update content as needed.\n\n6. **Feedback Loop**: Often, the results are looped back into the model's reasoningprocess, which can then adjust its line of questioning or actions based on the new data.\n\nThis system enhances the LLM's flexibility, making itapplicable to a wider range of real-world applications by utilizing up-to-date and specialized information through these tools.",  # noqa: E501
    },
]


@tool_eval()
def search_by_title_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools searching for objects by title."""
    suite = EvalSuite(
        name="Notion Search Tools Evaluation",
        system_message=(
            "You are an AI assistant that has access to the user's Notion workspace. "
            "You can take actions on the user's Notion workspace on behalf of the user."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    # Easy case
    suite.add_case(
        name="Search by title easy difficulty",
        user_message="Search for my page with the title 'Daily Standup'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_by_title,
                args={
                    "query": "Daily Standup",
                    "select": ObjectType.PAGE,
                },
            ),
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=0.6, similarity_threshold=0.95),
            BinaryCritic(critic_field="select", weight=0.4),
        ],
    )

    # Medium case
    suite.add_case(
        name="Search by title medium difficulty",
        user_message=(
            "so i was just thinking about LLMs and how to create an agent. "
            "I remember that tools are important for some reason. "
            "do i have a page or db about tool calling?"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_by_title,
                args={
                    "query": "tool calling",
                    "select": None,
                },
            ),
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=0.6),
            BinaryCritic(critic_field="select", weight=0.4),
        ],
    )

    # Hard case
    suite.add_case(
        name="Search by title hard difficulty",
        user_message=(
            "do i have any notes about any of those breakdown points? "
            "Actually, do I have any notes about the 2nd, 3rd, or 5th points?"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_by_title,
                args={
                    "query": "Tool Selection",
                    "select": None,
                },
            ),
            ExpectedToolCall(
                func=search_by_title,
                args={
                    "query": "Parameter Configuration",
                    "select": None,
                },
            ),
            ExpectedToolCall(
                func=search_by_title,
                args={
                    "query": "Result Integration",
                    "select": None,
                },
            ),
        ],
        critics=[
            SimilarityCritic(critic_field="query", weight=0.8),
            BinaryCritic(critic_field="select", weight=0.2),
        ],
        additional_messages=TOOL_CALLING_CONVERSATION,
    )

    return suite


@tool_eval()
def get_object_metadata_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting object metadata."""
    suite = EvalSuite(
        name="Notion Get Object Metadata Evaluation",
        system_message=(
            "You are an AI assistant that has access to the user's Notion workspace. "
            "You can take actions on the user's Notion workspace on behalf of the user."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    # Easy case
    suite.add_case(
        name="Get object metadata easy difficulty",
        user_message="Get any metadata about my page with the title 'Daily Standup'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_object_metadata,
                args={
                    "object_title": "Daily Standup",
                    "object_type": ObjectType.PAGE,
                },
            ),
        ],
        critics=[
            SimilarityCritic(critic_field="object_title", weight=0.8, similarity_threshold=0.95),
            BinaryCritic(critic_field="object_type", weight=0.2),
        ],
    )

    # Medium case
    suite.add_case(
        name="Get object metadata medium difficulty",
        user_message="Get the id, url, and last edited time of 'Daily Standup'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_object_metadata,
                args={
                    "object_title": "Daily Standup",
                    "object_type": None,
                },
            ),
        ],
        critics=[
            SimilarityCritic(critic_field="object_title", weight=0.8, similarity_threshold=0.95),
            BinaryCritic(critic_field="object_type", weight=0.2),
        ],
    )

    # Hard case
    suite.add_case(
        name="Get object metadata hard difficulty",
        user_message=(
            "oh I have page about that second point. "
            "This page here https://www.notion.so/be633bf1dfa0436db259571129a590e5. "
            "When was it created?"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_object_metadata,
                args={
                    "object_id": "be633bf1dfa0436db259571129a590e5",
                    "object_type": ObjectType.PAGE,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="object_id", weight=0.8),
            BinaryCritic(critic_field="object_type", weight=0.2),
        ],
        additional_messages=TOOL_CALLING_CONVERSATION,
    )

    return suite


@tool_eval()
def get_workspace_structure_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting the workspace structure."""
    suite = EvalSuite(
        name="Notion Get Workspace Structure Evaluation",
        system_message=(
            "You are an AI assistant that has access to the user's Notion workspace. "
            "You can take actions on the user's Notion workspace on behalf of the user."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    # Easy case
    suite.add_case(
        name="Get workspace structure easy difficulty",
        user_message="Get my workspace tree structure",
        expected_tool_calls=[
            ExpectedToolCall(func=get_workspace_structure, args={}),
        ],
    )

    # Medium case
    suite.add_case(
        name="Get workspace structure medium difficulty",
        user_message="I'm trying to figure out where my 'Daily Standup' page is. "
        "Can you help me find it?",
        expected_tool_calls=[
            ExpectedToolCall(func=get_workspace_structure, args={}),
        ],
    )

    # Hard case
    suite.add_case(
        name="Get workspace structure hard difficulty",
        user_message="list pages that are subpages of my 'Daily Standup' page",
        expected_tool_calls=[
            ExpectedToolCall(func=get_workspace_structure, args={}),
        ],
    )
    return suite



import json

from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_gmail
from arcade_gmail.enums import GmailReplyToWhom
from arcade_gmail.tools import (
    get_thread,
    list_emails_by_header,
    list_threads,
    reply_to_email,
    search_threads,
    send_email,
    write_draft_reply_email,
)
from arcade_gmail.utils import DateRange

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_gmail)


@tool_eval()
def gmail_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Gmail tools."""
    suite = EvalSuite(
        name="Gmail Tools Evaluation",
        system_message="You are an AI assistant that can send and manage emails using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Send email to user with clear username",
        user_message="Send a email to johndoe@example.com saying 'Hello, can we meet at 3 PM?'. CC his boss janedoe@example.com",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_email,
                args={
                    "subject": "Meeting Request",
                    "body": "Hello, can we meet at 3 PM?",
                    "recipient": "johndoe@example.com",
                    "cc": ["janedoe@example.com"],
                    "bcc": None,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="subject", weight=0.125),
            SimilarityCritic(critic_field="body", weight=0.25),
            BinaryCritic(critic_field="recipient", weight=0.25),
            BinaryCritic(critic_field="cc", weight=0.25),
            BinaryCritic(critic_field="bcc", weight=0.125),
        ],
    )

    suite.add_case(
        name="Simple list threads",
        user_message="Get 42 threads like right now i even wanna see the ones in my trash",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_threads,
                args={"max_results": 42, "include_spam_trash": True},
            )
        ],
        critics=[
            BinaryCritic(critic_field="max_results", weight=0.5),
            BinaryCritic(critic_field="include_spam_trash", weight=0.5),
        ],
    )

    history = [
        {"role": "user", "content": "list 1 thread"},
        {
            "role": "assistant",
            "content": "",
            "tool_calls": [
                {
                    "id": "call_X8V5Hw9iJ3wfB8WMZf8omAMi",
                    "type": "function",
                    "function": {"name": "Google_ListThreads", "arguments": '{"max_results":1}'},
                }
            ],
        },
        {
            "role": "tool",
            "content": '{"next_page_token":"10321400718999360131","num_threads":1,"threads":[{"historyId":"61691","id":"1934a8f8deccb749","snippet":"Hi Joe, I hope this email finds you well. Thank you for being a part of our community."}]}',
            "tool_call_id": "call_X8V5Hw9iJ3wfB8WMZf8omAMi",
            "name": "Google_ListThreads",
        },
        {
            "role": "assistant",
            "content": "Here is one email thread:\n\n- **Snippet:** Hi Joe, I hope this email finds you well. Thank you for being a part of our community.\n- **Thread ID:** 1934a8f8deccb749\n- **History ID:** 61691",
        },
    ]
    suite.add_case(
        name="List threads with history",
        user_message="Get the next 5 threads",
        additional_messages=history,
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_threads,
                args={
                    "max_results": 5,
                    "page_token": "10321400718999360131",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="max_results", weight=0.2),
            BinaryCritic(critic_field="page_token", weight=0.8),
        ],
    )

    suite.add_case(
        name="Search threads",
        user_message="Search for threads from johndoe@example.com to janedoe@example.com about that talk about 'Arcade AI' from yesterday",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_threads,
                args={
                    "sender": "johndoe@example.com",
                    "recipient": "janedoe@example.com",
                    "body": "Arcade AI",
                    "date_range": DateRange.YESTERDAY,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="sender", weight=0.25),
            BinaryCritic(critic_field="recipient", weight=0.25),
            SimilarityCritic(critic_field="body", weight=0.25),
            BinaryCritic(critic_field="date_range", weight=0.25),
        ],
    )

    suite.add_case(
        name="Get a thread by ID",
        user_message="Get the thread r-124325435467568867667878874565464564563523424323524235242412",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_thread,
                args={
                    "thread_id": "r-124325435467568867667878874565464564563523424323524235242412",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="thread_id", weight=1.0),
        ],
    )

    return suite


@tool_eval()
def gmail_reply_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Gmail reply tools."""
    suite = EvalSuite(
        name="Gmail Reply Tools Evaluation",
        system_message="You are an AI assistant that can send and manage emails using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    email_history = [
        {"role": "user", "content": "get the latest emails I received from johndoe@gmail.com"},
        {
            "role": "assistant",
            "content": "",
            "tool_calls": [
                {
                    "id": "call_jowMD7aB9sVPClOfvNof7Llu",
                    "type": "function",
                    "function": {
                        "name": "Google_ListEmailsByHeader",
                        "arguments": json.dumps({
                            "sender": "johndoe@gmail.com",
                            "max_results": 5,
                        }),
                    },
                }
            ],
        },
        {
            "role": "tool",
            "content": json.dumps({
                "emails": [
                    {
                        "body": "test 1",
                        "cc": "",
                        "date": "Tue, 11 Feb 2025 11:33:08 -0300",
                        "from": "John Doe <johndoe@gmail.com>",
                        "header_message_id": "<cierty475cty7245yxq@mail.gmail.com>",
                        "history_id": "123456",
                        "id": "q34759q435nv",
                        "in_reply_to": "",
                        "label_ids": ["INBOX"],
                        "references": "",
                        "reply_to": "",
                        "snippet": "test 1",
                        "subject": "test 1",
                        "thread_id": "345y6v3596",
                        "to": "myself@gmail.com",
                    },
                    {
                        "body": "test 2",
                        "cc": "",
                        "date": "Mon, 20 Jan 2025 13:04:42 -0800",
                        "from": "John Doe <johndoe@gmail.com>",
                        "header_message_id": "<28745ytvw8745ct4@mail.gmail.com>",
                        "history_id": "3456758",
                        "id": "9475tvy24578yx",
                        "in_reply_to": "",
                        "label_ids": [],
                        "references": "",
                        "reply_to": "",
                        "snippet": "test 2",
                        "subject": "test 2",
                        "thread_id": "249576v3496",
                        "to": "myself@gmail.com",
                    },
                ]
            }),
            "tool_call_id": "call_jowMD7aB9sVPClOfvNof7Llu",
            "name": "Google_ListEmailsByHeader",
        },
        {
            "role": "assistant",
            "content": "Here are the latest emails you received from johndoe@gmail.com:\n\n1. **Subject**: test 1\n   - **Date**: Tue, 11 Feb 2025 11:33:08 -0300\n   - **Snippet**: test 1\n\n2. **Subject**: test 2\n   - **Date**: Mon, 20 Jan 2025 13:04:42 -0800\n   - **Snippet**: test 2\n\nIf you need further details from any specific email, let me know!",
        },
    ]

    suite.add_case(
        name="Reply to an email",
        user_message="Reply to the email from johndoe@example.com about 'test 2' saying 'tested and working well'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=reply_to_email,
                args={
                    "reply_to_message_id": "9475tvy24578yx",
                    "body": "tested and working well",
                    "reply_to_whom": GmailReplyToWhom.ONLY_THE_SENDER.value,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="subject", weight=1 / 7),
            SimilarityCritic(critic_field="body", weight=1 / 7),
            BinaryCritic(critic_field="recipient", weight=1 / 7),
            BinaryCritic(critic_field="cc", weight=1 / 7),
            BinaryCritic(critic_field="bcc", weight=1 / 7),
            BinaryCritic(critic_field="reply_to_whom", weight=1 / 7),
            BinaryCritic(critic_field="reply_to_message_id", weight=1 / 7),
        ],
        additional_messages=email_history,
    )

    suite.add_case(
        name="Reply to an email with every recipient",
        user_message="Reply to every recipient in the email from johndoe@example.com about 'test 2' saying 'tested and working well'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=reply_to_email,
                args={
                    "reply_to_message_id": "9475tvy24578yx",
                    "body": "tested and working well",
                    "reply_to_whom": GmailReplyToWhom.EVERY_RECIPIENT.value,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="subject", weight=1 / 7),
            SimilarityCritic(critic_field="body", weight=1 / 7),
            BinaryCritic(critic_field="recipient", weight=1 / 7),
            BinaryCritic(critic_field="cc", weight=1 / 7),
            BinaryCritic(critic_field="bcc", weight=1 / 7),
            BinaryCritic(critic_field="reply_to_whom", weight=1 / 7),
            BinaryCritic(critic_field="reply_to_message_id", weight=1 / 7),
        ],
        additional_messages=email_history,
    )

    suite.add_case(
        name="Reply to an email with bcc",
        user_message="Reply to the email from johndoe@example.com about 'test 2' saying 'tested and working well' and send it to janedoe@example.com as bcc as well",
        expected_tool_calls=[
            ExpectedToolCall(
                func=reply_to_email,
                args={
                    "reply_to_message_id": "9475tvy24578yx",
                    "body": "tested and working well",
                    "bcc": ["janedoe@example.com"],
                    "reply_to_whom": GmailReplyToWhom.ONLY_THE_SENDER.value,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="subject", weight=1 / 7),
            SimilarityCritic(critic_field="body", weight=1 / 7),
            BinaryCritic(critic_field="recipient", weight=1 / 7),
            BinaryCritic(critic_field="cc", weight=1 / 7),
            BinaryCritic(critic_field="bcc", weight=1 / 7),
            BinaryCritic(critic_field="reply_to_whom", weight=1 / 7),
            BinaryCritic(critic_field="reply_to_message_id", weight=1 / 7),
        ],
        additional_messages=email_history,
    )

    suite.add_case(
        name="Write draft reply",
        user_message="Write a draft reply to the email from johndoe@example.com about 'test 2' saying 'tested and working well'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=write_draft_reply_email,
                args={
                    "reply_to_message_id": "9475tvy24578yx",
                    "body": "tested and working well",
                    "reply_to_whom": GmailReplyToWhom.ONLY_THE_SENDER.value,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="subject", weight=1 / 7),
            SimilarityCritic(critic_field="body", weight=1 / 7),
            BinaryCritic(critic_field="recipient", weight=1 / 7),
            BinaryCritic(critic_field="cc", weight=1 / 7),
            BinaryCritic(critic_field="bcc", weight=1 / 7),
            BinaryCritic(critic_field="reply_to_message_id", weight=1 / 7),
            BinaryCritic(critic_field="reply_to_whom", weight=1 / 7),
        ],
        additional_messages=email_history,
    )

    suite.add_case(
        name="Write draft reply to every recipient",
        user_message="Write a draft reply to every recipient in the email from johndoe@example.com about 'test 2' saying 'tested and working well'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=write_draft_reply_email,
                args={
                    "reply_to_message_id": "9475tvy24578yx",
                    "body": "tested and working well",
                    "reply_to_whom": GmailReplyToWhom.EVERY_RECIPIENT.value,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="subject", weight=1 / 7),
            SimilarityCritic(critic_field="body", weight=1 / 7),
            BinaryCritic(critic_field="recipient", weight=1 / 7),
            BinaryCritic(critic_field="cc", weight=1 / 7),
            BinaryCritic(critic_field="bcc", weight=1 / 7),
            BinaryCritic(critic_field="reply_to_whom", weight=0.125),
            BinaryCritic(critic_field="reply_to_message_id", weight=1 / 7),
        ],
        additional_messages=email_history,
    )

    return suite


@tool_eval()
def gmail_list_emails_by_header_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Gmail tools."""
    suite = EvalSuite(
        name="Gmail list_emails_by_header tool evaluation",
        system_message="You are an AI assistant that can send and manage emails using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="List emails by header using date-range",
        user_message="List all emails from johndoe@example.com to janedoe@example.com about 'Arcade AI' from yesterday",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_emails_by_header,
                args={
                    "sender": "johndoe@example.com",
                    "recipient": "janedoe@example.com",
                    "subject": "Arcade AI",
                    "date_range": DateRange.YESTERDAY.value,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="sender", weight=1 / 4),
            BinaryCritic(critic_field="recipient", weight=1 / 4),
            SimilarityCritic(critic_field="subject", weight=1 / 4),
            BinaryCritic(critic_field="date_range", weight=1 / 4),
        ],
    )

    suite.add_case(
        name="List emails by header using date-range",
        user_message="List all emails from johndoe@example.com to janedoe@example.com about 'Arcade AI' from the last month",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_emails_by_header,
                args={
                    "sender": "johndoe@example.com",
                    "recipient": "janedoe@example.com",
                    "subject": "Arcade AI",
                    "date_range": DateRange.LAST_MONTH.value,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="sender", weight=1 / 4),
            BinaryCritic(critic_field="recipient", weight=1 / 4),
            SimilarityCritic(critic_field="subject", weight=1 / 4),
            BinaryCritic(critic_field="date_range", weight=1 / 4),
        ],
    )

    return suite
